import os
import pathlib
import tempfile
from collections.abc import Sequence
from os import PathLike

import supriya.exceptions
from supriya.system import SupriyaValueObject

from ..io import PlayMemo
from .bases import ServerObject


class Buffer(ServerObject):
    """
    A buffer.

    ::

        >>> server = supriya.Server().boot()

    ::

        >>> buffer_ = supriya.realtime.Buffer()
        >>> buffer_
        <- Buffer: ???>

    ::

        >>> buffer_ = buffer_.allocate(server, frame_count=8192)
        >>> server.sync()
        <Server: udp://127.0.0.1:57110, 8i8o>

    ::

        >>> buffer_
        <+ Buffer: 0, 1ch, 8192>

    ::

        >>> buffer_ = buffer_.free()
        >>> buffer_
        <- Buffer: ???>

    """

    ### CLASS VARIABLES ###

    __slots__ = ("_buffer_group", "_buffer_id", "_buffer_id_was_set_manually")

    ### INITIALIZER ###

    def __init__(self, buffer_group_or_index=None):
        ServerObject.__init__(self)
        buffer_group = None
        buffer_id = None
        self._buffer_id_was_set_manually = False
        if buffer_group_or_index is not None:
            self._buffer_id_was_set_manually = True
            if isinstance(buffer_group_or_index, BufferGroup):
                buffer_group = buffer_group_or_index
            elif isinstance(buffer_group_or_index, int):
                buffer_id = int(buffer_group_or_index)
        self._buffer_group = buffer_group
        self._buffer_id = buffer_id

    ### SPECIAL METHODS ###

    def __float__(self):
        """
        Gets float representation of buffer.

        ::

            >>> server = supriya.Server().boot()
            >>> buffer_one = supriya.realtime.Buffer().allocate(server, frame_count=8)
            >>> buffer_two = supriya.realtime.Buffer().allocate(server, frame_count=8)

        ::

            >>> buffer_one
            <+ Buffer: 0, 1ch, 8>

        ::

            >>> float(buffer_one)
            0.0

        ::

            >>> buffer_two
            <+ Buffer: 1, 1ch, 8>

        ::

            >>> float(buffer_two)
            1.0

        ::

            >>> buffer_one = buffer_one.free()
            >>> buffer_two = buffer_two.free()

        ::

            >>> buffer_three = supriya.realtime.Buffer()
            >>> float(buffer_three)
            Traceback (most recent call last):
            ...
            supriya.exceptions.BufferNotAllocated

        Returns float.
        """
        if not self.is_allocated:
            raise supriya.exceptions.BufferNotAllocated
        return float(self.buffer_id)

    def __int__(self):
        """
        Gets integer representation of buffer.

        ::

            >>> server = supriya.Server().boot()
            >>> buffer_one = supriya.realtime.Buffer().allocate(server, frame_count=8)
            >>> buffer_two = supriya.realtime.Buffer().allocate(server, frame_count=8)

        ::

            >>> buffer_one
            <+ Buffer: 0, 1ch, 8>

        ::

            >>> int(buffer_one)
            0

        ::

            >>> buffer_two
            <+ Buffer: 1, 1ch, 8>

        ::

            >>> int(buffer_two)
            1

        ::

            >>> buffer_one = buffer_one.free()
            >>> buffer_two = buffer_two.free()

        ::

            >>> buffer_three = supriya.realtime.Buffer()
            >>> int(buffer_three)
            Traceback (most recent call last):
            ...
            supriya.exceptions.BufferNotAllocated

        Returns integer.
        """
        if not self.is_allocated:
            raise supriya.exceptions.BufferNotAllocated
        return int(self.buffer_id)

    def __plot__(self):
        import librosa

        with tempfile.TemporaryDirectory() as temp_directory:
            file_path = pathlib.Path(temp_directory) / "tmp.wav"
            self.write(file_path=file_path, header_format="wav", sample_format="int32")
            return librosa.load(file_path, mono=False, sr=None)

    def __render__(self, **kwargs) -> PlayMemo:
        with tempfile.TemporaryDirectory() as temp_directory:
            file_path = pathlib.Path(temp_directory) / "tmp.wav"
            self.write(file_path=file_path, header_format="wav", sample_format="int32")
            return PlayMemo.from_path(file_path)

    def __repr__(self):
        """
        Gets interpreter representation of buffer.

        ::

            >>> server = supriya.Server().boot()
            >>> buffer_ = supriya.realtime.Buffer()
            >>> repr(buffer_)
            '<- Buffer: ???>'

        ::

            >>> buffer_ = buffer_.allocate(server, frame_count=8)
            >>> repr(buffer_)
            '<+ Buffer: 0, 1ch, 8>'

        ::

            >>> buffer_ = buffer_.free()
            >>> repr(buffer_)
            '<- Buffer: ???>'

        Returns string.
        """
        if not self.is_allocated:
            return f"<- {type(self).__name__}: ???>"
        return f"<+ {type(self).__name__}: {self.buffer_id}, {self.channel_count}ch, {self.frame_count}>"

    ### PRIVATE METHODS ###

    def _allocate_buffer_id(self):
        if self.buffer_id is None:
            buffer_id = self.server.buffer_allocator.allocate(1)
            if buffer_id is None:
                ServerObject.free(self)
                raise ValueError
            self._buffer_id = buffer_id

    def _register_with_local_server(self):
        if self.buffer_id not in self.server._buffers:
            self.server._buffers[self.buffer_id] = set()
        self.server._buffers[self.buffer_id].add(self)
        self.server._get_buffer_proxy(self.buffer_id)

    def _register_with_remote_server(
        self,
        channel_count=None,
        frame_count=None,
        file_path=None,
        channel_indices=None,
        starting_frame=None,
    ):
        import supriya.commands

        on_done = supriya.commands.BufferQueryRequest(buffer_ids=(self.buffer_id,))
        if file_path and channel_indices is not None:
            if not isinstance(channel_indices, Sequence):
                channel_indices = (channel_indices,)
            channel_indices = tuple(channel_indices)
            assert all(0 <= _ for _ in channel_indices)
            request = supriya.commands.BufferAllocateReadChannelRequest(
                buffer_id=self.buffer_id,
                channel_indices=channel_indices,
                file_path=file_path,
                frame_count=frame_count,
                starting_frame=starting_frame,
                callback=on_done,
            )
        elif file_path:
            request = supriya.commands.BufferAllocateReadRequest(
                buffer_id=self.buffer_id,
                file_path=file_path,
                frame_count=frame_count,
                starting_frame=starting_frame,
                callback=on_done,
            )
        else:
            request = supriya.commands.BufferAllocateRequest(
                buffer_id=self.buffer_id,
                frame_count=frame_count,
                channel_count=channel_count,
                callback=on_done,
            )
        return request

    def _unregister_with_local_server(self):
        buffer_id = self.buffer_id
        buffers = self.server._buffers[buffer_id]
        buffers.remove(self)
        if not buffers:
            del self.server._buffers[buffer_id]
        return buffer_id

    def _unregister_with_remote_server(self, buffer_id):
        import supriya.commands

        on_done = supriya.commands.BufferQueryRequest(buffer_ids=(buffer_id,))
        request = supriya.commands.BufferFreeRequest(
            buffer_id=buffer_id, callback=on_done
        )
        return request

    ### PUBLIC METHODS ###

    def allocate(self, server, channel_count=1, frame_count=1, *, sync=True):
        """
        Allocates buffer on `server`.

        ::

            >>> server = supriya.Server().boot()
            >>> buffer_one = supriya.realtime.Buffer().allocate(server)
            >>> buffer_one.query()
            BufferInfoResponse(
                items=(
                    Item(buffer_id=0, frame_count=1, channel_count=1, sample_rate=4...00.0),
                ),
            )

        ::

            >>> buffer_two = supriya.realtime.Buffer().allocate(
            ...     server,
            ...     frame_count=16,
            ... )
            >>> buffer_two.query()
            BufferInfoResponse(
                items=(
                    Item(buffer_id=1, frame_count=16, channel_count=1, sample_rate=4...00.0),
                ),
            )

        ::

            >>> buffer_three = supriya.realtime.Buffer().allocate(
            ...     server,
            ...     channel_count=2,
            ...     frame_count=32,
            ... )
            >>> buffer_three.query()
            BufferInfoResponse(
                items=(
                    Item(buffer_id=2, frame_count=32, channel_count=2, sample_rate=4...00.0),
                ),
            )

        ::

            >>> buffer_three.allocate(server)
            Traceback (most recent call last):
            ...
            supriya.exceptions.BufferAlreadyAllocated

        ::

            >>> buffer_one = buffer_one.free()
            >>> buffer_two = buffer_two.free()
            >>> buffer_three = buffer_three.free()

        Returns buffer.
        """
        if self.is_allocated:
            raise supriya.exceptions.BufferAlreadyAllocated
        channel_count = int(channel_count)
        frame_count = int(frame_count)
        if channel_count < 1:
            raise ValueError(channel_count)
        if frame_count < 1:
            raise ValueError(frame_count)
        try:
            ServerObject.allocate(self, server=server)
            self._allocate_buffer_id()
            self._register_with_local_server()
            request = self._register_with_remote_server(
                channel_count=channel_count, frame_count=frame_count
            )
            request.communicate(server=self.server, sync=sync)
        except Exception:
            self.free()
        return self

    def allocate_from_file(
        self,
        server,
        file_path: PathLike,
        *,
        channel_indices=None,
        callback=None,
        frame_count=None,
        starting_frame=None,
        sync=True,
    ):
        """
        Allocates buffer on `server` with contents read from `file_path`.

        ::

            >>> server = supriya.Server().boot()
            >>> buffer_one = supriya.realtime.Buffer().allocate_from_file(
            ...     server,
            ...     supriya.system.Assets["audio/pulse_44100sr_16bit_octo.wav"],
            ... )
            >>> buffer_one.query()
            BufferInfoResponse(
                items=(
                    Item(buffer_id=0, frame_count=8, channel_count=8, sample_rate=4...00.0),
                ),
            )

        ::

            >>> buffer_two = supriya.realtime.Buffer().allocate_from_file(
            ...     server,
            ...     supriya.system.Assets["audio/pulse_44100sr_16bit_octo.wav"],
            ...     channel_indices=(3, 4),
            ...     frame_count=4,
            ...     starting_frame=1,
            ...     sync=True,
            ... )
            >>> buffer_two.query()
            BufferInfoResponse(
                items=(
                    Item(buffer_id=1, frame_count=4, channel_count=2, sample_rate=4...00.0),
                ),
            )

        ::

            >>> for frame_id in range(buffer_two.frame_count):
            ...     buffer_two.get_frames(frame_id).as_dict()
            ...
            OrderedDict([(0, (0.0, 0.0))])
            OrderedDict([(2, (0.0, 0.0))])
            OrderedDict([(4, (0.999969482421875, 0.0))])
            OrderedDict([(6, (0.0, 0.999969482421875))])

        ::

            >>> buffer_two.allocate(server)
            Traceback (most recent call last):
            ...
            supriya.exceptions.BufferAlreadyAllocated

        ::

            >>> buffer_one = buffer_one.free()
            >>> buffer_two = buffer_two.free()

        Returns buffer.
        """
        if self.is_allocated:
            raise supriya.exceptions.BufferAlreadyAllocated
        try:
            ServerObject.allocate(self, server=server)
            self._allocate_buffer_id()
            self._register_with_local_server()
            request = self._register_with_remote_server(
                frame_count=frame_count,
                channel_indices=channel_indices,
                file_path=file_path,
                starting_frame=starting_frame,
            )
            request.communicate(server=self.server, sync=sync)
        except Exception:
            ServerObject.allocate(self, server=server)
        return self

    def close(self, *, sync=True):
        """
        Closes buffer, if it was open during a read or write process by
        the DiskIn or DiskOut UGens.

        ::

            >>> server = supriya.Server().boot()
            >>> buffer_ = supriya.realtime.Buffer().allocate(
            ...     server, channel_count=8, frame_count=8
            ... )
            >>> buffer_.read(
            ...     supriya.system.Assets["audio/pulse_44100sr_16bit_octo.wav"],
            ...     leave_open=True,
            ... )
            >>> buffer_.close()
            >>> buffer_ = buffer_.free()

        ::

            >>> buffer_.close()
            Traceback (most recent call last):
            ...
            supriya.exceptions.BufferNotAllocated

        Returns none.
        """
        import supriya.commands

        if not self.is_allocated:
            raise supriya.exceptions.BufferNotAllocated
        request = supriya.commands.BufferCloseRequest(buffer_id=self.buffer_id)
        request.communicate(server=self.server, sync=sync)

    def copy(
        self,
        target_buffer_id,
        *,
        frame_count=None,
        source_starting_frame=None,
        target_starting_frame=None,
        sync=True,
    ):
        """
        Copies data in this buffer into another buffer.

        ::

            >>> server = supriya.Server().boot()
            >>> buffer_one = supriya.realtime.Buffer().allocate(server, frame_count=4)
            >>> buffer_two = supriya.realtime.Buffer().allocate(server, frame_count=4)
            >>> buffer_one.fill((0, 4, 0.5))
            >>> buffer_one.copy(target_buffer_id=buffer_two)
            >>> buffer_two.get_contiguous((0, 4)).as_dict()
            OrderedDict([(0, (0.5, 0.5, 0.5, 0.5))])

        ::

            >>> buffer_one = buffer_one.free()
            >>> buffer_two = buffer_two.free()

        ::

            >>> buffer_one.copy(target_buffer_id=666)
            Traceback (most recent call last):
            ...
            supriya.exceptions.BufferNotAllocated

        Returns none.
        """
        import supriya.commands

        if not self.is_allocated:
            raise supriya.exceptions.BufferNotAllocated
        request = supriya.commands.BufferCopyRequest(
            frame_count=frame_count,
            source_buffer_id=self.buffer_id,
            source_starting_frame=source_starting_frame,
            target_buffer_id=target_buffer_id,
            target_starting_frame=target_starting_frame,
        )
        request.communicate(server=self.server, sync=sync)

    def fill(self, *index_count_value_triples, sync=True):
        """
        Fills contiguous blocks of samples with values.

        ::

            >>> server = supriya.Server().boot()
            >>> buffer_ = supriya.realtime.Buffer().allocate(
            ...     server,
            ...     frame_count=8,
            ...     sync=True,
            ... )
            >>> buffer_.fill((0, 2, 0.5), (3, 3, 1.0))
            >>> buffer_.get_contiguous((0, 8)).as_dict()
            OrderedDict([(0, (0.5, 0.5, 0.0, 1.0, 1.0, 1.0, 0.0, 0.0))])

        ::

            >>> buffer_ = buffer_.free()

        ::

            >>> buffer_.fill()
            Traceback (most recent call last):
            ...
            supriya.exceptions.BufferNotAllocated

        Returns none.
        """
        import supriya.commands

        if not self.is_allocated:
            raise supriya.exceptions.BufferNotAllocated
        request = supriya.commands.BufferFillRequest(
            buffer_id=self.buffer_id,
            index_count_value_triples=index_count_value_triples,
        )
        request.communicate(server=self.server, sync=sync)

    def free(self) -> "Buffer":
        """
        Frees buffer.
        """
        if not self.is_allocated:
            raise supriya.exceptions.BufferNotAllocated
        buffer_id = self._unregister_with_local_server()
        request = self._unregister_with_remote_server(buffer_id)
        if self.server.is_running:
            request.communicate(server=self.server, sync=False)
        if not self._buffer_id_was_set_manually:
            self.server.buffer_allocator.free(self.buffer_id)
        self._buffer_id = None
        ServerObject.free(self)
        return self

    def fill_via_chebyshev(
        self,
        amplitudes,
        *,
        as_wavetable=False,
        should_normalize=True,
        should_clear_first=True,
        sync=True,
    ):
        """
        Fills buffer with Chebyshev polynomial.

        ::

            >>> server = supriya.Server().boot()
            >>> buffer_ = server.add_buffer(1, 512)
            >>> buffer_.fill_via_chebyshev(
            ...     [1, 0.5, 0.25],
            ...     as_wavetable=False,
            ... )
            >>> supriya.plot(buffer_)  # doctest: +SKIP

        ::

            >>> buffer_ = buffer_.free()

        ::

            >>> buffer_.fill_via_chebyshev([1, 0.5, 0.25])
            Traceback (most recent call last):
            ...
            supriya.exceptions.BufferNotAllocated

        Returns none.
        """
        import supriya.commands

        if not self.is_allocated:
            raise supriya.exceptions.BufferNotAllocated
        request = supriya.commands.BufferGenerateRequest.chebyshev(
            amplitudes=amplitudes,
            as_wavetable=as_wavetable,
            buffer_id=self.buffer_id,
            should_clear_first=should_clear_first,
            should_normalize=should_normalize,
        )
        request.communicate(server=self.server, sync=sync)

    def fill_via_sine_1(
        self,
        amplitudes,
        *,
        as_wavetable=False,
        should_clear_first=True,
        should_normalize=True,
        sync=True,
    ):
        """
        Fills buffer with sum of sinusoids via `/b_gen sine1`.

        ::

            >>> server = supriya.Server().boot()
            >>> buffer_ = server.add_buffer(1, 512)
            >>> buffer_.fill_via_sine_1(
            ...     [1, 1, 1],
            ...     as_wavetable=False,
            ... )
            >>> supriya.plot(buffer_)  # doctest: +SKIP

        ::

            >>> buffer_ = buffer_.free()

        ::

            >>> buffer_.fill_via_sine_1([1, 1, 1])
            Traceback (most recent call last):
            ...
            supriya.exceptions.BufferNotAllocated

        Returns none.
        """
        import supriya.commands

        if not self.is_allocated:
            raise supriya.exceptions.BufferNotAllocated
        request = supriya.commands.BufferGenerateRequest.sine1(
            amplitudes=amplitudes,
            as_wavetable=as_wavetable,
            buffer_id=self.buffer_id,
            should_clear_first=should_clear_first,
            should_normalize=should_normalize,
        )
        request.communicate(server=self.server, sync=sync)

    def fill_via_sine_2(
        self,
        amplitudes,
        frequencies,
        *,
        as_wavetable=False,
        should_clear_first=True,
        should_normalize=True,
        sync=True,
    ):
        """
        Fills buffer with sum of sinusoids via `/b_gen sine2`.

        ::

            >>> server = supriya.Server().boot()
            >>> buffer_ = server.add_buffer(1, 512)
            >>> buffer_.fill_via_sine_2(
            ...     frequencies=[1, 2, 4],
            ...     amplitudes=[1, 0.5, 0.25],
            ...     as_wavetable=False,
            ... )
            >>> supriya.plot(buffer_)  # doctest: +SKIP

        ::

            >>> buffer_ = buffer_.free()

        ::

            >>> buffer_.fill_via_sine_2(
            ...     frequencies=[1, 2, 4],
            ...     amplitudes=[1, 0.5, 0.25],
            ... )
            Traceback (most recent call last):
            ...
            supriya.exceptions.BufferNotAllocated

        Returns none.
        """
        import supriya.commands

        if not self.is_allocated:
            raise supriya.exceptions.BufferNotAllocated
        request = supriya.commands.BufferGenerateRequest.sine2(
            amplitudes=amplitudes,
            frequencies=frequencies,
            as_wavetable=as_wavetable,
            buffer_id=self.buffer_id,
            should_clear_first=should_clear_first,
            should_normalize=should_normalize,
        )
        request.communicate(server=self.server, sync=sync)

    def fill_via_sine_3(
        self,
        amplitudes,
        frequencies,
        phases,
        *,
        as_wavetable=False,
        should_clear_first=True,
        should_normalize=True,
        sync=True,
    ):
        """
        Fills buffer with sum of sinusoids via `/b_gen sine3`.

        ::

            >>> server = supriya.Server().boot()
            >>> buffer_ = server.add_buffer(1, 512)
            >>> buffer_.fill_via_sine_3(
            ...     frequencies=[1, 2, 3],
            ...     amplitudes=[1, 0.5, 0.25],
            ...     phases=[0, 0.5, 0],
            ...     as_wavetable=False,
            ... )
            >>> supriya.plot(buffer_)  # doctest: +SKIP

        ::

            >>> buffer_ = buffer_.free()

        ::

            >>> buffer_.fill_via_sine_3(
            ...     frequencies=[1, 2, 3],
            ...     amplitudes=[1, 0.5, 0.25],
            ...     phases=[0, 0.5, 0],
            ... )
            Traceback (most recent call last):
            ...
            supriya.exceptions.BufferNotAllocated

        Returns none.
        """
        import supriya.commands

        if not self.is_allocated:
            raise supriya.exceptions.BufferNotAllocated
        request = supriya.commands.BufferGenerateRequest.sine3(
            amplitudes=amplitudes,
            frequencies=frequencies,
            phases=phases,
            as_wavetable=as_wavetable,
            buffer_id=self.buffer_id,
            should_clear_first=should_clear_first,
            should_normalize=should_normalize,
        )
        request.communicate(server=self.server, sync=sync)

    def get(self, *indices):
        """
        Gets sample values at `indices`.

        ::

            >>> server = supriya.Server().boot()
            >>> buffer_ = supriya.realtime.Buffer().allocate(
            ...     server,
            ...     frame_count=4,
            ...     sync=True,
            ... )
            >>> response = buffer_.get(1, 2)
            >>> response.as_dict()
            OrderedDict([(1, 0.0), (2, 0.0)])

        ::

            >>> buffer_ = buffer_.free()

        ::

            >>> buffer_.get()
            Traceback (most recent call last):
            ...
            supriya.exceptions.BufferNotAllocated

        Returns buffer-set response.
        """
        import supriya.commands

        if not self.is_allocated:
            raise supriya.exceptions.BufferNotAllocated
        request = supriya.commands.BufferGetRequest(buffer_id=self, indices=indices)
        response = request.communicate(server=self.server)
        if isinstance(response, supriya.commands.FailResponse):
            raise IndexError("Index out of range.")
        return response

    def get_contiguous(self, *index_count_pairs):
        """
        Gets contiguous sample values.

        ::

            >>> server = supriya.Server().boot()
            >>> buffer_ = supriya.realtime.Buffer().allocate(
            ...     server,
            ...     frame_count=4,
            ...     sync=True,
            ... )
            >>> response = buffer_.get_contiguous((0, 2), (1, 3))
            >>> response.as_dict()
            OrderedDict([(0, (0.0, 0.0)), (1, (0.0, 0.0, 0.0))])

        ::

            >>> buffer_ = buffer_.free()

        ::

            >>> buffer_.get_contiguous()
            Traceback (most recent call last):
            ...
            supriya.exceptions.BufferNotAllocated

        Returns buffer-set-contiguous response.
        """
        import supriya.commands

        if not self.is_allocated:
            raise supriya.exceptions.BufferNotAllocated
        request = supriya.commands.BufferGetContiguousRequest(
            buffer_id=self, index_count_pairs=index_count_pairs
        )
        response = request.communicate(server=self.server)
        if isinstance(response, supriya.commands.FailResponse):
            raise IndexError("Index out of range.")
        return response

    def get_frames(self, *frame_ids, completion_callback=None):
        """
        Gets frames at `frame_ids`.

        ::

            >>> server = supriya.Server().boot()
            >>> buffer_ = supriya.realtime.Buffer().allocate_from_file(
            ...     server,
            ...     supriya.system.Assets["audio/pulse_44100sr_16bit_octo.wav"],
            ... )
            >>> for frame_id in range(buffer_.frame_count):
            ...     buffer_.get_frames(frame_id).as_dict()
            ...
            OrderedDict([(0, (0.999969482421875, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0))])
            OrderedDict([(8, (0.0, 0.999969482421875, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0))])
            OrderedDict([(16, (0.0, 0.0, 0.999969482421875, 0.0, 0.0, 0.0, 0.0, 0.0))])
            OrderedDict([(24, (0.0, 0.0, 0.0, 0.999969482421875, 0.0, 0.0, 0.0, 0.0))])
            OrderedDict([(32, (0.0, 0.0, 0.0, 0.0, 0.999969482421875, 0.0, 0.0, 0.0))])
            OrderedDict([(40, (0.0, 0.0, 0.0, 0.0, 0.0, 0.999969482421875, 0.0, 0.0))])
            OrderedDict([(48, (0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.999969482421875, 0.0))])
            OrderedDict([(56, (0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.999969482421875))])

        ::

            >>> buffer_ = buffer_.free()

        ::

            >>> buffer_.get_frames()
            Traceback (most recent call last):
            ...
            supriya.exceptions.BufferNotAllocated

        Returns buffer-set-contiguous response.
        """
        if not self.is_allocated:
            raise supriya.exceptions.BufferNotAllocated
        if isinstance(frame_ids, int):
            frame_ids = [frame_ids]
        index_count_pairs = [
            (frame_id * self.channel_count, self.channel_count)
            for frame_id in frame_ids
        ]
        response = self.get_contiguous(*index_count_pairs)
        return response

    def normalize(self, *, as_wavetable=None, new_maximum=1.0, sync=True):
        request = supriya.commands.BufferNormalizeRequest(
            as_wavetable=as_wavetable, buffer_id=self, new_maximum=new_maximum
        )
        request.communicate(server=self.server, sync=sync)

    def query(self):
        """
        Queries buffer.

        ::

            >>> server = supriya.Server().boot()
            >>> buffer_ = supriya.realtime.Buffer().allocate(
            ...     server, channel_count=2, frame_count=16
            ... )
            >>> buffer_.query()
            BufferInfoResponse(
                items=(
                    Item(buffer_id=0, frame_count=16, channel_count=2, sample_rate=4...00.0),
                ),
            )

        ::

            >>> buffer_ = buffer_.free()

        ::

            >>> buffer_.query()
            Traceback (most recent call last):
            ...
            supriya.exceptions.BufferNotAllocated

        Returns buffer-info response.
        """
        import supriya.commands

        if not self.is_allocated:
            raise supriya.exceptions.BufferNotAllocated
        buffer_ids = [self.buffer_id]
        request = supriya.commands.BufferQueryRequest(buffer_ids=buffer_ids)
        response = request.communicate(server=self.server)
        return response

    def read(
        self,
        file_path,
        *,
        channel_indices=None,
        callback=None,
        frame_count=None,
        leave_open=None,
        starting_frame_in_buffer=None,
        starting_frame_in_file=None,
        sync=True,
    ):
        """
        Reads contents of `file_path` into buffer.

        ::

            >>> server = supriya.Server().boot()
            >>> buffer_ = supriya.realtime.Buffer().allocate(
            ...     server, channel_count=2, frame_count=8
            ... )
            >>> for frame_id in range(buffer_.frame_count):
            ...     buffer_.get_frames(frame_id).as_dict()
            ...
            OrderedDict([(0, (0.0, 0.0))])
            OrderedDict([(2, (0.0, 0.0))])
            OrderedDict([(4, (0.0, 0.0))])
            OrderedDict([(6, (0.0, 0.0))])
            OrderedDict([(8, (0.0, 0.0))])
            OrderedDict([(10, (0.0, 0.0))])
            OrderedDict([(12, (0.0, 0.0))])
            OrderedDict([(14, (0.0, 0.0))])

        ::

            >>> file_path = supriya.system.Assets["audio/pulse_44100sr_16bit_octo.wav"]
            >>> buffer_.read(file_path, channel_indices=(0, 1))

        ::

            >>> for frame_id in range(buffer_.frame_count):
            ...     buffer_.get_frames(frame_id).as_dict()
            ...
            OrderedDict([(0, (0.999969482421875, 0.0))])
            OrderedDict([(2, (0.0, 0.999969482421875))])
            OrderedDict([(4, (0.0, 0.0))])
            OrderedDict([(6, (0.0, 0.0))])
            OrderedDict([(8, (0.0, 0.0))])
            OrderedDict([(10, (0.0, 0.0))])
            OrderedDict([(12, (0.0, 0.0))])
            OrderedDict([(14, (0.0, 0.0))])

        ::

            >>> buffer_ = buffer_.free()

        ::

            >>> buffer_.read(file_path)
            Traceback (most recent call last):
            ...
            supriya.exceptions.BufferNotAllocated

        Returns none.
        """
        import supriya.commands

        if not self.is_allocated:
            raise supriya.exceptions.BufferNotAllocated
        on_done = supriya.commands.BufferQueryRequest(buffer_ids=(self.buffer_id,))
        if channel_indices is not None:
            request = supriya.commands.BufferReadChannelRequest(
                buffer_id=self.buffer_id,
                channel_indices=channel_indices,
                callback=on_done,
                file_path=file_path,
                frame_count=frame_count,
                leave_open=leave_open,
                starting_frame_in_buffer=starting_frame_in_buffer,
                starting_frame_in_file=starting_frame_in_file,
            )
        else:
            request = supriya.commands.BufferReadRequest(
                buffer_id=self.buffer_id,
                callback=on_done,
                file_path=file_path,
                frame_count=frame_count,
                leave_open=leave_open,
                starting_frame_in_buffer=starting_frame_in_buffer,
                starting_frame_in_file=starting_frame_in_file,
            )
        request.communicate(server=self.server, sync=sync)

    def set(self, *index_value_pairs, sync=False):
        """
        Sets samples.

        ::

            >>> server = supriya.Server().boot()
            >>> buffer_ = supriya.realtime.Buffer().allocate(server, frame_count=8)

        ::

            >>> buffer_.set((0, 0.25), (1, 0.5), (4, 0.75), (5, 1.0))
            >>> buffer_.get_contiguous((0, 8)).as_dict()[0]
            (0.25, 0.5, 0.0, 0.0, 0.75, 1.0, 0.0, 0.0)

        ::

            >>> buffer_ = buffer_.free()

        ::

            >>> buffer_.set()
            Traceback (most recent call last):
            ...
            supriya.exceptions.BufferNotAllocated

        Returns none.
        """
        import supriya.commands

        if not self.is_allocated:
            raise supriya.exceptions.BufferNotAllocated
        request = supriya.commands.BufferSetRequest(
            buffer_id=self, index_value_pairs=index_value_pairs
        )
        request.communicate(server=self.server, sync=sync)

    def set_contiguous(self, *index_values_pairs, sync=False):
        """
        Sets contiguous blocks of samples.

        ::

            >>> server = supriya.Server().boot()
            >>> buffer_ = supriya.realtime.Buffer().allocate(server, frame_count=8)

        ::

            >>> buffer_.set_contiguous((1, [1, 2, 3]), (4, [-3, 2, -1]))
            >>> buffer_.get_contiguous((0, 8)).as_dict()[0]
            (0.0, 1.0, 2.0, 3.0, -3.0, 2.0, -1.0, 0.0)

        ::

            >>> buffer_ = buffer_.free()

        ::

            >>> buffer_.set_contiguous()
            Traceback (most recent call last):
            ...
            supriya.exceptions.BufferNotAllocated

        Returns none.
        """
        import supriya.commands

        if not self.is_allocated:
            raise supriya.exceptions.BufferNotAllocated
        request = supriya.commands.BufferSetContiguousRequest(
            buffer_id=self, index_values_pairs=index_values_pairs
        )
        request.communicate(server=self.server, sync=sync)

    def write(
        self,
        file_path,
        *,
        callback=None,
        frame_count=None,
        header_format="aiff",
        leave_open=False,
        sample_format="int24",
        starting_frame=None,
        sync=True,
    ):
        """
        Writes buffer to disk.

        ::

            >>> server = supriya.Server().boot()
            >>> buffer_one = supriya.realtime.Buffer().allocate_from_file(
            ...     server,
            ...     supriya.system.Assets["audio/pulse_44100sr_16bit_octo.wav"],
            ...     channel_indices=(0,),
            ... )
            >>> buffer_one.get_contiguous((0, 8)).as_dict()[0]
            (0.999969482421875, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0)

        ::

            >>> import os
            >>> file_path = os.path.expanduser("~")
            >>> file_path = os.path.join(file_path, "temp.wav")
            >>> if os.path.exists(file_path):
            ...     os.remove(file_path)

        ::

            >>> buffer_one.write(
            ...     file_path,
            ...     header_format="wav",
            ... )

        ::

            >>> buffer_two = supriya.realtime.Buffer().allocate_from_file(server, file_path)
            >>> buffer_two.get_contiguous((0, 8)).as_dict()[0]
            (0.999969482421875, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0)

        ::

            >>> os.remove(file_path)
            >>> buffer_one = buffer_one.free()
            >>> buffer_two = buffer_two.free()

        ::

            >>> buffer_one.write(file_path=file_path)
            Traceback (most recent call last):
            ...
            supriya.exceptions.BufferNotAllocated

        Returns none.
        """
        import supriya.commands

        if not self.is_allocated:
            raise supriya.exceptions.BufferNotAllocated
        request = supriya.commands.BufferWriteRequest(
            buffer_id=self.buffer_id,
            callback=callback,
            file_path=file_path,
            frame_count=frame_count,
            header_format=header_format,
            leave_open=leave_open,
            sample_format=sample_format,
            starting_frame=starting_frame,
        )
        request.communicate(server=self.server, sync=sync)

    def zero(self, *, callback=None, sync=True):
        """
        Zero all samples in buffer.

        ::

            >>> server = supriya.Server().boot()
            >>> buffer_ = supriya.realtime.Buffer().allocate(server, frame_count=8)

        ::

            >>> buffer_.set_contiguous(
            ...     (0, (1, 2, 3, 4, 5, 6, 7, 8)),
            ...     sync=True,
            ... )

        ::

            >>> buffer_.get_contiguous((0, 8)).as_dict()
            OrderedDict([(0, (1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0))])

        ::

            >>> buffer_.zero()

        ::

            >>> buffer_.get_contiguous((0, 8)).as_dict()
            OrderedDict([(0, (0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0))])

        ::

            >>> buffer_ = buffer_.free()

        ::

            >>> buffer_.zero()
            Traceback (most recent call last):
            ...
            supriya.exceptions.BufferNotAllocated

        Returns none.
        """
        import supriya.commands

        if not self.is_allocated:
            raise supriya.exceptions.BufferNotAllocated
        request = supriya.commands.BufferZeroRequest(
            buffer_id=self.buffer_id, callback=callback
        )
        request.communicate(server=self.server, sync=sync)

    ### PUBLIC PROPERTIES ###

    @property
    def buffer_group(self):
        """
        Gets buffer group.

        ::

            >>> server = supriya.Server().boot()
            >>> buffer_one = supriya.realtime.Buffer().allocate(server, frame_count=8)
            >>> buffer_one.buffer_group is None
            True

        ::

            >>> buffer_group = supriya.realtime.BufferGroup(buffer_count=1)
            >>> buffer_group.allocate(server, frame_count=8)
            <+ BufferGroup{1}: 1>

        ::

            >>> buffer_two = buffer_group[0]
            >>> buffer_two.buffer_group is buffer_group
            True

        ::

            >>> buffer_one = buffer_one.free()
            >>> buffer_one.buffer_group is None
            True

        ::

            >>> buffer_group = buffer_group.free()
            >>> buffer_two.buffer_group is buffer_group
            True

        Returns BufferGroup or none.
        """
        return self._buffer_group

    @property
    def buffer_id(self):
        """
        Gets buffer id.

        ::

            >>> server = supriya.Server().boot()
            >>> buffer_one = supriya.realtime.Buffer()
            >>> buffer_one.buffer_id is None
            True

        ::

            >>> buffer_group = supriya.realtime.BufferGroup(buffer_count=4)
            >>> for buffer_ in buffer_group:
            ...     print(buffer_.buffer_id)
            ...
            None
            None
            None
            None

        ::

            >>> buffer_one = buffer_one.allocate(server, frame_count=8)
            >>> buffer_one.buffer_id
            0

        ::

            >>> buffer_group.allocate(server, frame_count=8)
            <+ BufferGroup{4}: 1>

        ::

            >>> for buffer_ in buffer_group:
            ...     buffer_.buffer_id
            ...
            1
            2
            3
            4

        ::

            >>> buffer_one = buffer_one.free()
            >>> buffer_one.buffer_id is None
            True

        ::

            >>> buffer_group = buffer_group.free()
            >>> for buffer_ in buffer_group:
            ...     print(buffer_.buffer_id)
            ...
            None
            None
            None
            None

        Returns integer or none.
        """
        if self._buffer_group is not None:
            if self._buffer_group.buffer_id is not None:
                group_id = self._buffer_group.buffer_id
                index = self._buffer_group.index(self)
                buffer_id = group_id + index
                return buffer_id
        return self._buffer_id

    @property
    def channel_count(self):
        """
        Gets channel count.

        ::

            >>> server = supriya.Server().boot()
            >>> buffer_one = supriya.realtime.Buffer().allocate(server, frame_count=8)

        ::

            >>> buffer_two = supriya.realtime.Buffer().allocate(
            ...     server, channel_count=4, frame_count=8
            ... )

        ::

            >>> buffer_one.channel_count
            1

        ::

            >>> buffer_two.channel_count
            4

        ::

            >>> buffer_one = buffer_one.free()
            >>> buffer_two = buffer_two.free()

        Returns integer.
        """
        if self.is_allocated:
            proxy = self.server._buffer_proxies[self.buffer_id]
            return proxy.channel_count
        return 0

    @property
    def duration_in_seconds(self):
        """
        Gets duration in seconds.

        ::

            >>> server = supriya.Server().boot()
            >>> buffer_ = supriya.realtime.Buffer()
            >>> buffer_.duration_in_seconds
            0.0

        ::

            >>> buffer_ = buffer_.allocate(server, frame_count=44100)
            >>> buffer_.duration_in_seconds == buffer_.frame_count / buffer_.sample_rate
            True

        ::

            >>> buffer_ = buffer_.free()
            >>> buffer_.duration_in_seconds
            0.0

        Returns float.
        """
        if self.is_allocated:
            proxy = self.server._buffer_proxies[self.buffer_id]
            return proxy.duration_in_seconds
        return 0.0

    @property
    def frame_count(self):
        """
        Gets frame count.

        ::

            >>> server = supriya.Server().boot()
            >>> buffer_ = supriya.realtime.Buffer()
            >>> buffer_.frame_count
            0

        ::

            >>> buffer_ = buffer_.allocate(server, frame_count=512)
            >>> buffer_.frame_count
            512

        ::

            >>> buffer_ = buffer_.free()
            >>> buffer_.frame_count
            0

        Returns integer.
        """
        if self.is_allocated:
            proxy = self.server._buffer_proxies[self.buffer_id]
            return proxy.frame_count
        return 0

    @property
    def sample_count(self):
        """
        Gets sample count.

        ::

            >>> server = supriya.Server().boot()
            >>> buffer_one = supriya.realtime.Buffer().allocate(server, frame_count=16)
            >>> buffer_two = supriya.realtime.Buffer().allocate(
            ...     server,
            ...     channel_count=2,
            ...     frame_count=16,
            ... )
            >>> buffer_three = supriya.realtime.Buffer().allocate(
            ...     server,
            ...     channel_count=8,
            ...     frame_count=16,
            ... )

        ::

            >>> buffer_one.sample_count
            16

        ::

            >>> buffer_two.sample_count
            32

        ::

            >>> buffer_three.sample_count
            128

        ::

            >>> buffer_one = buffer_one.free()
            >>> buffer_two = buffer_two.free()
            >>> buffer_three = buffer_three.free()

        ::

            >>> buffer_one.sample_count
            0

        ::

            >>> buffer_two.sample_count
            0

        ::

            >>> buffer_three.sample_count
            0

        Returns integer.
        """
        if self.is_allocated:
            proxy = self.server._buffer_proxies[self.buffer_id]
            return proxy.sample_count
        return 0

    @property
    def sample_rate(self):
        """
        Gets sample-rate.

        ::

            >>> server = supriya.Server().boot()
            >>> buffer_ = supriya.realtime.Buffer()
            >>> buffer_.sample_rate
            0

        ::

            >>> buffer_ = buffer_.allocate(server, frame_count=8)
            >>> buffer_.sample_rate
            4...00.0

        ::

            >>> buffer_ = buffer_.free()
            >>> buffer_.sample_rate
            0

        Returns float.
        """
        if self.is_allocated:
            proxy = self.server._buffer_proxies[self.buffer_id]
            return proxy.sample_rate
        return 0

    @property
    def is_allocated(self):
        """
        Is true if buffer is allocated. Otherwise false.

        ::

            >>> server = supriya.Server().boot()
            >>> buffer_ = supriya.realtime.Buffer()
            >>> buffer_.is_allocated
            False

        ::

            >>> buffer_ = buffer_.allocate(server, frame_count=8)
            >>> buffer_.is_allocated
            True

        ::

            >>> buffer_ = buffer_.free()
            >>> buffer_.is_allocated
            False

        Returns boolean
        """
        if self.buffer_group is not None:
            return self.buffer_group.is_allocated
        return self.server is not None

    @property
    def server(self):
        """
        Gets associated server.

        ::

            >>> server = supriya.Server().boot()
            >>> buffer_ = supriya.realtime.Buffer()
            >>> buffer_.server is None
            True

        ::

            >>> buffer_ = buffer_.allocate(server, frame_count=8)
            >>> buffer_.server is server
            True

        ::

            >>> buffer_ = buffer_.free()
            >>> buffer_.server is None
            True

        Returns server or none.
        """
        if self.buffer_group is not None:
            return self.buffer_group.server
        return self._server


class BufferGroup(ServerObject):
    """
    A buffer group.

    ::

        >>> server = supriya.Server().boot()

    ::

        >>> buffer_group = supriya.realtime.BufferGroup(buffer_count=4)
        >>> buffer_group
        <- BufferGroup{4}: ???>

    ::

        >>> buffer_group.allocate(
        ...     server,
        ...     frame_count=8192,
        ...     sync=True,
        ... )
        <+ BufferGroup{4}: 0>

    ::

        >>> buffer_group.free()
        <- BufferGroup{4}: ???>

    """

    ### CLASS VARIABLES ###

    __slots__ = ("_buffer_id", "_buffers")

    ### INITIALIZER ###

    def __init__(self, buffer_count=1):
        ServerObject.__init__(self)
        self._buffer_id = None
        buffer_count = int(buffer_count)
        assert 0 < buffer_count
        self._buffers = tuple(
            Buffer(buffer_group_or_index=self) for _ in range(buffer_count)
        )

    ### SPECIAL METHODS ###

    def __contains__(self, item):
        return self.buffers.__contains__(item)

    def __float__(self):
        return float(self.buffer_id)

    def __getitem__(self, index):
        """
        Gets buffer at `index`.

        Returns buffer.
        """
        return self._buffers[index]

    def __int__(self):
        return int(self.buffer_id)

    def __iter__(self):
        return iter(self.buffers)

    def __len__(self):
        """
        Gets length of buffer group.

        Returns integer.
        """
        return len(self._buffers)

    def __repr__(self):
        """
        Gets interpreter representation of buffer group.

        Returns string.
        """
        buffer_id = self.buffer_id
        if buffer_id is None:
            buffer_id = "???"
        string = "<{} {}{{{}}}: {}>".format(
            "+" if self.is_allocated else "-", type(self).__name__, len(self), buffer_id
        )
        return string

    ### PRIVATE METHODS ###

    def _register_with_local_server(self, server):
        ServerObject.allocate(self, server=server)
        allocator = self.server.buffer_allocator
        buffer_id = allocator.allocate(len(self))
        if buffer_id is None:
            ServerObject.free(self)
            raise ValueError
        self._buffer_id = buffer_id
        for buffer_ in self:
            buffer_._register_with_local_server()
        return buffer_id

    ### PUBLIC METHODS ###

    def allocate(self, server, channel_count=1, frame_count=None, sync=True):
        """
        Allocates buffer group.

        Returns buffer group.
        """
        if self.is_allocated:
            return supriya.exceptions.BufferAlreadyAllocated
        channel_count = int(channel_count)
        frame_count = int(frame_count)
        assert 0 < channel_count
        assert 0 < frame_count
        self._register_with_local_server(server)
        requests = []
        for buffer_ in self:
            requests.append(
                buffer_._register_with_remote_server(
                    channel_count=channel_count, frame_count=frame_count
                )
            )
        supriya.commands.RequestBundle(contents=requests).communicate(
            server=server, sync=sync
        )
        return self

    def free(self) -> "BufferGroup":
        """
        Frees all buffers in buffer group.
        """
        if not self.is_allocated:
            raise supriya.exceptions.BufferNotAllocated
        for buffer_ in self:
            buffer_.free()
        buffer_id = self.buffer_id
        self._buffer_id = None
        self.server.buffer_allocator.free(buffer_id)
        ServerObject.free(self)
        return self

    def index(self, item):
        return self.buffers.index(item)

    @staticmethod
    def from_file_paths(server, file_paths, sync=True):
        """
        Create a buffer group from `file_paths`.

        ::

            >>> file_paths = supriya.Assets["audio/*mono_1s*"]
            >>> len(file_paths)
            4

        ::

            >>> server = supriya.Server().boot()
            >>> buffer_group = supriya.realtime.BufferGroup.from_file_paths(server, file_paths)

        ::

            >>> for buffer_ in buffer_group:
            ...     buffer_
            ...
            <+ Buffer: 0, 1ch, 44100>
            <+ Buffer: 1, 1ch, 44100>
            <+ Buffer: 2, 1ch, 44100>
            <+ Buffer: 3, 1ch, 44100>

        Returns buffer group.
        """
        for file_path in file_paths:
            assert os.path.exists(file_path)
        buffer_group = BufferGroup(buffer_count=len(file_paths))
        buffer_group._register_with_local_server(server)
        requests = []
        for buffer_, file_path in zip(buffer_group.buffers, file_paths):
            request = buffer_._register_with_remote_server(file_path=file_path)
            requests.append(request)
        supriya.commands.RequestBundle(contents=requests).communicate(server, sync=sync)
        return buffer_group

    def zero(self):
        """
        Analogous to SuperCollider's Buffer.zero.
        """
        raise NotImplementedError

    ### PUBLIC PROPERTIES ###

    @property
    def buffer_id(self):
        """
        Gets initial buffer id.

        Returns integer or none.
        """
        return self._buffer_id

    @property
    def buffers(self):
        """
        Gets associated buffers.

        Returns tuple or buffers.
        """
        return self._buffers

    @property
    def is_allocated(self):
        """
        Is true when buffer group is allocated. Otherwise false.

        Returns boolean.
        """
        return self.server is not None


class BufferProxy(SupriyaValueObject):
    """
    A buffer proxy.

    Acts as a singleton reference to a buffer on the server, tracking the state
    of a single buffer id and responding to `/b_info` messages. Multiple Buffer
    instances reference a single BufferProxy.

    BufferProxy instances are created internally by the server, and should be
    treated as an implementation detail.

    ::

        >>> server = supriya.Server()
        >>> buffer_proxy = supriya.realtime.BufferProxy(
        ...     buffer_id=0,
        ...     server=server,
        ...     channel_count=2,
        ...     frame_count=441,
        ...     sample_rate=44100,
        ... )
        >>> buffer_proxy
        BufferProxy(
            buffer_id=0,
            channel_count=2,
            frame_count=441,
            sample_rate=44100,
            server=<Server: offline>,
        )

    """

    ### CLASS VARIABLES ###

    __slots__ = (
        "_buffer_id",
        "_channel_count",
        "_frame_count",
        "_sample_rate",
        "_server",
    )

    ### INITIALIZER ###

    def __init__(
        self, buffer_id=None, channel_count=0, frame_count=0, sample_rate=0, server=None
    ):
        import supriya.realtime

        buffer_id = int(buffer_id)
        assert 0 <= buffer_id
        assert isinstance(server, supriya.realtime.Server)
        self._buffer_id = int(buffer_id)
        self._channel_count = int(channel_count)
        self._frame_count = int(frame_count)
        self._sample_rate = int(sample_rate)
        self._server = server

    ### SPECIAL METHODS ###

    def __float__(self):
        """
        Gets float representation of buffer proxy.

        ::

            >>> server = supriya.Server()
            >>> buffer_proxy = supriya.realtime.BufferProxy(
            ...     buffer_id=0,
            ...     server=server,
            ...     channel_count=2,
            ...     frame_count=441,
            ...     sample_rate=44100,
            ... )
            >>> float(buffer_proxy)
            0.0

        Returns float.
        """
        return float(self.buffer_id)

    def __int__(self):
        """
        Gets integer representation of buffer proxy.

        ::

            >>> server = supriya.Server()
            >>> buffer_proxy = supriya.realtime.BufferProxy(
            ...     buffer_id=0,
            ...     server=server,
            ...     channel_count=2,
            ...     frame_count=441,
            ...     sample_rate=44100,
            ... )
            >>> int(buffer_proxy)
            0

        Returns integer.
        """
        return int(self.buffer_id)

    ### PRIVATE METHODS ###

    def _handle_response(self, response):
        """
        Updates buffer proxy with buffer-info response.

        ::

            >>> server = supriya.Server()
            >>> a_buffer = supriya.realtime.BufferProxy(
            ...     buffer_id=23,
            ...     channel_count=1,
            ...     frame_count=256,
            ...     sample_rate=44100,
            ...     server=server,
            ... )
            >>> a_buffer
            BufferProxy(
                buffer_id=23,
                channel_count=1,
                frame_count=256,
                sample_rate=44100,
                server=<Server: offline>,
            )

        ::

            >>> response_item = supriya.commands.BufferInfoResponse.Item(
            ...     buffer_id=23,
            ...     channel_count=2,
            ...     frame_count=512,
            ...     sample_rate=44100,
            ... )

        ::

            >>> a_buffer._handle_response(response_item)
            >>> a_buffer
            BufferProxy(
                buffer_id=23,
                channel_count=2,
                frame_count=512,
                sample_rate=44100,
                server=<Server: offline>,
            )

        Returns none.
        """
        import supriya.commands

        if isinstance(response, supriya.commands.BufferInfoResponse.Item):
            assert response.buffer_id == self.buffer_id
            self._channel_count = response.channel_count
            self._frame_count = response.frame_count
            self._sample_rate = response.sample_rate

    ### PUBLIC PROPERTIES ###

    @property
    def buffer_id(self):
        """
        Gets buffer id of buffer proxy.

        ::

            >>> server = supriya.Server()
            >>> buffer_proxy = supriya.realtime.BufferProxy(
            ...     buffer_id=0,
            ...     server=server,
            ...     channel_count=2,
            ...     frame_count=441,
            ...     sample_rate=44100,
            ... )
            >>> buffer_proxy.buffer_id
            0

        Returns integer.
        """
        return self._buffer_id

    @property
    def channel_count(self):
        """
        Gets channel count of buffer proxy.

        ::

            >>> server = supriya.Server()
            >>> buffer_proxy = supriya.realtime.BufferProxy(
            ...     buffer_id=0,
            ...     server=server,
            ...     channel_count=2,
            ...     frame_count=441,
            ...     sample_rate=44100,
            ... )
            >>> buffer_proxy.channel_count
            2

        Returns integer.
        """
        return self._channel_count

    @property
    def duration_in_seconds(self):
        """
        Gets duration in seconds of buffer proxy.

        ::

            >>> server = supriya.Server()
            >>> buffer_proxy = supriya.realtime.BufferProxy(
            ...     buffer_id=0,
            ...     server=server,
            ...     channel_count=2,
            ...     frame_count=441,
            ...     sample_rate=44100,
            ... )
            >>> buffer_proxy.duration_in_seconds
            0.01

        Returns float.
        """
        return float(self._frame_count) / float(self.sample_rate)

    @property
    def frame_count(self):
        """
        Gets frame count of buffer proxy.

        ::

            >>> server = supriya.Server()
            >>> buffer_proxy = supriya.realtime.BufferProxy(
            ...     buffer_id=0,
            ...     server=server,
            ...     channel_count=2,
            ...     frame_count=441,
            ...     sample_rate=44100,
            ... )
            >>> buffer_proxy.frame_count
            441

        Returns integer.
        """
        return self._frame_count

    @property
    def sample_count(self):
        """
        Gets sample count of buffer proxy.

        ::

            >>> server = supriya.Server()
            >>> buffer_proxy = supriya.realtime.BufferProxy(
            ...     buffer_id=0,
            ...     server=server,
            ...     channel_count=2,
            ...     frame_count=441,
            ...     sample_rate=44100,
            ... )
            >>> buffer_proxy.sample_count
            882

        Returns integer.
        """
        return self._channel_count * self._frame_count

    @property
    def sample_rate(self):
        """
        Gets sample-rate of buffer proxy.

        ::

            >>> server = supriya.Server()
            >>> buffer_proxy = supriya.realtime.BufferProxy(
            ...     buffer_id=0,
            ...     server=server,
            ...     channel_count=2,
            ...     frame_count=441,
            ...     sample_rate=44100,
            ... )
            >>> buffer_proxy.sample_rate
            44100

        Returns integer.
        """
        return self._sample_rate

    @property
    def server(self):
        """
        Gets server of buffer proxy.

        ::

            >>> server = supriya.Server()
            >>> buffer_proxy = supriya.realtime.BufferProxy(
            ...     buffer_id=0,
            ...     server=server,
            ...     channel_count=2,
            ...     frame_count=441,
            ...     sample_rate=44100,
            ... )
            >>> buffer_proxy.server
            <Server: offline>

        Returns server.
        """
        return self._server



import supriya.synthdefs
import supriya.ugens

__all__ = []


def _build_link_audio_synthdef(channel_count):
    r"""
    SynthDef("system_link_audio_" ++ i, {
        arg out=0, in=16, vol=1, level=1, lag=0.05, doneAction=2;
        var env = EnvGate(doneAction:doneAction, curve:'sin') * Lag.kr(vol * level, lag);
        Out.ar(out, InFeedback.ar(in, i) * env);
    }, [\kr, \kr, \kr, \kr, \kr, \ir]).add;
    """
    name = "system_link_audio_{}".format(channel_count)
    builder = supriya.synthdefs.SynthDefBuilder(
        name=name, out=0, in_=16, gate=1, fade_time=0.02, done_action=2
    )
    with builder:
        start_value = builder["fade_time"] <= 0
        envelope = supriya.synthdefs.Envelope(
            amplitudes=[start_value, 1.0, 0.0],
            durations=[1.0, 1.0],
            curves=[supriya.EnvelopeShape.SINE, -supriya.EnvelopeShape.SINE],
            release_node=1.0,
        )
        envelope = supriya.ugens.EnvGen.kr(
            done_action=builder["done_action"],
            envelope=envelope,
            gate=builder["gate"],
            time_scale=builder["fade_time"],
        )
        source = supriya.ugens.InFeedback.ar(
            bus=builder["in_"], channel_count=channel_count
        )
        supriya.ugens.Out.ar(bus=builder["out"], source=source * envelope)
    globals()[name] = builder.build()
    __all__.append(name)


def _build_link_control_synthdef(channel_count):
    r"""
    SynthDef("system_link_control_" ++ i, {
        arg out=0, in=16, doneAction=2;
        var env = EnvGate(doneAction:doneAction, curve:'lin');
        Out.kr(out, In.kr(in, i) * env);
    }, [\kr, \kr, \ir]).add;
    """
    name = "system_link_control_{}".format(channel_count)
    builder = supriya.synthdefs.SynthDefBuilder(
        name=name, out=0, in_=16, gate=1, fade_time=0.02, done_action=2
    )
    with builder:
        start_value = builder["fade_time"] <= 0
        envelope = supriya.synthdefs.Envelope(
            amplitudes=[start_value, 1.0, 0.0],
            durations=[1.0, 1.0],
            curves=supriya.EnvelopeShape.LINEAR,
            release_node=1.0,
        )
        envelope = supriya.ugens.EnvGen.kr(
            done_action=builder["done_action"],
            envelope=envelope,
            gate=builder["gate"],
            time_scale=builder["fade_time"],
        )
        input_ = supriya.ugens.In.kr(bus=builder["in_"], channel_count=channel_count)
        supriya.ugens.Out.kr(bus=builder["out"], source=input_ * envelope)
    globals()[name] = builder.build()
    __all__.append(name)


for i in range(1, 17):
    _build_link_audio_synthdef(i)
    _build_link_control_synthdef(i)



import collections

from supriya import CalculationRate
from supriya.synthdefs import UGen, WidthFirstUGen


class PV_ChainUGen(WidthFirstUGen):
    """
    Abstract base class for all phase-vocoder-chain unit generators.
    """

    ### INITIALIZER ###

    def __init__(self, **kwargs):
        calculation_rate = CalculationRate.CONTROL
        WidthFirstUGen.__init__(self, calculation_rate=calculation_rate, **kwargs)

    ### PUBLIC PROPERTIES ###

    @property
    def fft_size(self):
        """
        Gets FFT size as UGen input.

        Returns ugen input.
        """
        return self.pv_chain.fft_size


class FFT(PV_ChainUGen):
    """
    A fast Fourier transform.

    ::

        >>> buffer_id = supriya.ugens.LocalBuf(2048)
        >>> source = supriya.ugens.SoundIn.ar(bus=0)
        >>> fft = supriya.ugens.FFT(
        ...     active=1,
        ...     buffer_id=buffer_id,
        ...     hop=0.5,
        ...     source=source,
        ...     window_size=0,
        ...     window_type=0,
        ... )
        >>> fft
        FFT.kr()

    """

    ### CLASS VARIABLES ###

    _ordered_input_names = collections.OrderedDict(
        [
            ("buffer_id", None),
            ("source", None),
            ("hop", 0.5),
            ("window_type", 0),
            ("active", 1),
            ("window_size", 0),
        ]
    )

    ### INITIALIZER ###

    def __init__(
        self,
        buffer_id=None,
        source=None,
        active=1,
        hop=0.5,
        window_size=0,
        window_type=0,
    ):
        import supriya.ugens

        if buffer_id is None:
            buffer_size = window_size or 2048
            buffer_id = supriya.ugens.LocalBuf(buffer_size)
        PV_ChainUGen.__init__(
            self,
            active=active,
            buffer_id=buffer_id,
            hop=hop,
            source=source,
            window_size=window_size,
            window_type=window_type,
        )

    ### PUBLIC PROPERTIES ###

    @property
    def fft_size(self):
        """
        Gets FFT size as UGen input.

        Returns ugen input.
        """
        import supriya.ugens

        return supriya.ugens.BufFrames.ir(self.buffer_id)


class IFFT(WidthFirstUGen):
    """
    An inverse fast Fourier transform.

    ::

        >>> pv_chain = supriya.ugens.LocalBuf(2048)
        >>> ifft = supriya.ugens.IFFT.ar(
        ...     pv_chain=pv_chain,
        ...     window_size=0,
        ...     window_type=0,
        ... )
        >>> ifft
        IFFT.ar()

    """

    _ordered_input_names = collections.OrderedDict(
        [("pv_chain", None), ("window_type", 0), ("window_size", 0)]
    )
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)


class PV_Add(PV_ChainUGen):
    """
    Complex addition.

    ::

        >>> pv_chain_a = supriya.ugens.FFT(
        ...     source=supriya.ugens.WhiteNoise.ar(),
        ... )
        >>> pv_chain_b = supriya.ugens.FFT(
        ...     source=supriya.ugens.LFSaw.ar(),
        ... )
        >>> pv_add = supriya.ugens.PV_Add.new(
        ...     pv_chain_a=pv_chain_a,
        ...     pv_chain_b=pv_chain_b,
        ... )
        >>> pv_add
        PV_Add.kr()

    """

    _ordered_input_names = collections.OrderedDict(
        [("pv_chain_a", None), ("pv_chain_b", None)]
    )


class PV_BinScramble(PV_ChainUGen):
    """
    Scrambles bins.

    ::

        >>> pv_chain = supriya.ugens.FFT(
        ...     source=supriya.ugens.WhiteNoise.ar(),
        ... )
        >>> pv_bin_scramble = supriya.ugens.PV_BinScramble.new(
        ...     pv_chain=pv_chain,
        ...     trigger=0,
        ...     width=0.2,
        ...     wipe=0,
        ... )
        >>> pv_bin_scramble
        PV_BinScramble.kr()

    """

    _ordered_input_names = collections.OrderedDict(
        [("pv_chain", None), ("wipe", 0), ("width", 0.2), ("trigger", 0)]
    )


class PV_BinShift(PV_ChainUGen):
    """
    Shifts and stretches bin positions.

    ::

        >>> pv_chain = supriya.ugens.FFT(
        ...     source=supriya.ugens.WhiteNoise.ar(),
        ... )
        >>> pv_bin_shift = supriya.ugens.PV_BinShift.new(
        ...     pv_chain=pv_chain,
        ...     interpolate=0,
        ...     shift=0,
        ...     stretch=1,
        ... )
        >>> pv_bin_shift
        PV_BinShift.kr()

    """

    _ordered_input_names = collections.OrderedDict(
        [("pv_chain", None), ("stretch", 1), ("shift", 0), ("interpolate", 0)]
    )


class PV_BinWipe(PV_ChainUGen):
    """
    Copies low bins from one input and the high bins of the other.

    ::

        >>> pv_chain_a = supriya.ugens.FFT(
        ...     source=supriya.ugens.WhiteNoise.ar(),
        ... )
        >>> pv_chain_b = supriya.ugens.FFT.new(
        ...     source=supriya.ugens.LFSaw.ar(),
        ... )
        >>> pv_bin_wipe = supriya.ugens.PV_BinWipe.new(
        ...     pv_chain_a=pv_chain_a,
        ...     pv_chain_b=pv_chain_b,
        ...     wipe=0,
        ... )
        >>> pv_bin_wipe
        PV_BinWipe.kr()

    """

    _ordered_input_names = collections.OrderedDict(
        [("pv_chain_a", None), ("pv_chain_b", None), ("wipe", 0)]
    )


class PV_BrickWall(PV_ChainUGen):
    """
    Zeros bins.

    - If wipe == 0 then there is no effect.
    - If wipe > 0 then it acts like a high pass filter, clearing bins from the
      bottom up.
    - If wipe < 0 then it acts like a low pass filter, clearing bins from the
      top down.

    ::

        >>> pv_chain = supriya.ugens.FFT(
        ...     source=supriya.ugens.WhiteNoise.ar(),
        ... )
        >>> pv_brick_wall = supriya.ugens.PV_BrickWall.new(
        ...     pv_chain=pv_chain,
        ...     wipe=0,
        ... )
        >>> pv_brick_wall
        PV_BrickWall.kr()

    """

    _ordered_input_names = collections.OrderedDict([("pv_chain", None), ("wipe", 0)])


class PV_ConformalMap(PV_ChainUGen):
    """
    Complex plane attack.

    ::

        >>> pv_chain = supriya.ugens.FFT(
        ...     source=supriya.ugens.WhiteNoise.ar(),
        ... )
        >>> pv_conformal_map = supriya.ugens.PV_ConformalMap.new(
        ...     aimag=0,
        ...     areal=0,
        ...     pv_chain=pv_chain,
        ... )
        >>> pv_conformal_map
        PV_ConformalMap.kr()

    """

    _ordered_input_names = collections.OrderedDict(
        [("pv_chain", None), ("areal", 0), ("aimag", 0)]
    )


class PV_Conj(PV_ChainUGen):
    """
    Complex conjugate.

    ::

        >>> pv_chain = supriya.ugens.FFT(
        ...     source=supriya.ugens.WhiteNoise.ar(),
        ... )
        >>> pv_conj = supriya.ugens.PV_Conj.new(
        ...     pv_chain=pv_chain,
        ... )
        >>> pv_conj
        PV_Conj.kr()

    """

    _ordered_input_names = collections.OrderedDict([("pv_chain", None)])


class PV_Copy(PV_ChainUGen):
    """
    Copies an FFT buffer.

    ::

        >>> pv_chain_a = supriya.ugens.FFT(
        ...     source=supriya.ugens.WhiteNoise.ar(),
        ... )
        >>> pv_chain_b = supriya.ugens.FFT(
        ...     source=supriya.ugens.LFSaw.ar(),
        ... )
        >>> pv_copy = supriya.ugens.PV_Copy.new(
        ...     pv_chain_a=pv_chain_a,
        ...     pv_chain_b=pv_chain_b,
        ... )
        >>> pv_copy
        PV_Copy.kr()

    """

    _ordered_input_names = collections.OrderedDict(
        [("pv_chain_a", None), ("pv_chain_b", None)]
    )


class PV_CopyPhase(PV_ChainUGen):
    """
    Copies magnitudes and phases.

    ::

        >>> pv_chain_a = supriya.ugens.FFT(
        ...     source=supriya.ugens.WhiteNoise.ar(),
        ... )
        >>> pv_chain_b = supriya.ugens.FFT(
        ...     source=supriya.ugens.LFSaw.ar(),
        ... )
        >>> pv_copy_phase = supriya.ugens.PV_CopyPhase.new(
        ...     pv_chain_a=pv_chain_a,
        ...     pv_chain_b=pv_chain_b,
        ... )
        >>> pv_copy_phase
        PV_CopyPhase.kr()

    """

    _ordered_input_names = collections.OrderedDict(
        [("pv_chain_a", None), ("pv_chain_b", None)]
    )


class PV_Diffuser(PV_ChainUGen):
    """
    Shifts phases randomly.

    ::

        >>> pv_chain = supriya.ugens.FFT(
        ...     source=supriya.ugens.WhiteNoise.ar(),
        ... )
        >>> pv_diffuser = supriya.ugens.PV_Diffuser.new(
        ...     pv_chain=pv_chain,
        ...     trigger=0,
        ... )
        >>> pv_diffuser
        PV_Diffuser.kr()

    """

    _ordered_input_names = collections.OrderedDict([("pv_chain", None), ("trigger", 0)])


class PV_Div(PV_ChainUGen):
    """
    Complex division.

    ::

        >>> pv_chain_a = supriya.ugens.FFT(
        ...     source=supriya.ugens.WhiteNoise.ar(),
        ... )
        >>> pv_chain_b = supriya.ugens.FFT(
        ...     source=supriya.ugens.LFSaw.ar(),
        ... )
        >>> pv_div = supriya.ugens.PV_Div.new(
        ...     pv_chain_a=pv_chain_a,
        ...     pv_chain_b=pv_chain_b,
        ... )
        >>> pv_div
        PV_Div.kr()

    """

    _ordered_input_names = collections.OrderedDict(
        [("pv_chain_a", None), ("pv_chain_b", None)]
    )


class PV_HainsworthFoote(PV_ChainUGen):
    """
    A FFT onset detector.

    ::

        >>> pv_chain = supriya.ugens.FFT(
        ...     source=supriya.ugens.WhiteNoise.ar(),
        ... )
        >>> pv_hainsworth_foote = supriya.ugens.PV_HainsworthFoote.new(
        ...     pv_chain=pv_chain,
        ...     propf=0,
        ...     proph=0,
        ...     threshold=1,
        ...     waittime=0.04,
        ... )
        >>> pv_hainsworth_foote
        PV_HainsworthFoote.kr()

    """

    _ordered_input_names = collections.OrderedDict(
        [
            ("pv_chain", None),
            ("proph", 0),
            ("propf", 0),
            ("threshold", 1),
            ("waittime", 0.04),
        ]
    )


class PV_JensenAndersen(PV_ChainUGen):
    """
    A FFT feature detector for onset detection.

    ::

        >>> pv_chain = supriya.ugens.FFT(
        ...     source=supriya.ugens.WhiteNoise.ar(),
        ... )
        >>> pv_jensen_andersen = supriya.ugens.PV_JensenAndersen.new(
        ...     pv_chain=pv_chain,
        ...     prophfc=0.25,
        ...     prophfe=0.25,
        ...     propsc=0.25,
        ...     propsf=0.25,
        ...     threshold=1,
        ...     waittime=0.04,
        ... )
        >>> pv_jensen_andersen
        PV_JensenAndersen.kr()

    """

    _ordered_input_names = collections.OrderedDict(
        [
            ("pv_chain", None),
            ("propsc", 0.25),
            ("prophfe", 0.25),
            ("prophfc", 0.25),
            ("propsf", 0.25),
            ("threshold", 1),
            ("waittime", 0.04),
        ]
    )


class PV_LocalMax(PV_ChainUGen):
    """
    Passes bins which are local maxima.

    ::

        >>> pv_chain = supriya.ugens.FFT(
        ...     source=supriya.ugens.WhiteNoise.ar(),
        ... )
        >>> pv_local_max = supriya.ugens.PV_LocalMax.new(
        ...     pv_chain=pv_chain,
        ...     threshold=0,
        ... )
        >>> pv_local_max
        PV_LocalMax.kr()

    """

    _ordered_input_names = collections.OrderedDict(
        [("pv_chain", None), ("threshold", 0)]
    )


class PV_MagAbove(PV_ChainUGen):
    """
    Passes magnitudes above threshold.

    ::

        >>> pv_chain = supriya.ugens.FFT(
        ...     source=supriya.ugens.WhiteNoise.ar(),
        ... )
        >>> pv_mag_above = supriya.ugens.PV_MagAbove.new(
        ...     pv_chain=pv_chain,
        ...     threshold=0,
        ... )
        >>> pv_mag_above
        PV_MagAbove.kr()

    """

    _ordered_input_names = collections.OrderedDict(
        [("pv_chain", None), ("threshold", 0)]
    )


class PV_MagBelow(PV_MagAbove):
    """
    Passes magnitudes below threshold.

    ::

        >>> pv_chain = supriya.ugens.FFT(
        ...     source=supriya.ugens.WhiteNoise.ar(),
        ... )
        >>> pv_mag_below = supriya.ugens.PV_MagBelow.new(
        ...     pv_chain=pv_chain,
        ...     threshold=0,
        ... )
        >>> pv_mag_below
        PV_MagBelow.kr()

    """

    _ordered_input_names = collections.OrderedDict(
        [("pv_chain", None), ("threshold", 0)]
    )


class PV_MagClip(PV_MagAbove):
    """
    Clips magnitudes.

    ::

        >>> pv_chain = supriya.ugens.FFT(
        ...     source=supriya.ugens.WhiteNoise.ar(),
        ... )
        >>> pv_mag_clip = supriya.ugens.PV_MagClip.new(
        ...     pv_chain=pv_chain,
        ...     threshold=0,
        ... )
        >>> pv_mag_clip
        PV_MagClip.kr()

    """

    _ordered_input_names = collections.OrderedDict(
        [("pv_chain", None), ("threshold", 0)]
    )


class PV_MagDiv(PV_ChainUGen):
    """
    Divides magnitudes.

    ::

        >>> pv_chain_a = supriya.ugens.FFT(
        ...     source=supriya.ugens.WhiteNoise.ar(),
        ... )
        >>> pv_chain_b = supriya.ugens.FFT(
        ...     source=supriya.ugens.LFSaw.ar(),
        ... )
        >>> pv_mag_div = supriya.ugens.PV_MagDiv.new(
        ...     pv_chain_a=pv_chain_a,
        ...     pv_chain_b=pv_chain_b,
        ...     zeroed=0.0001,
        ... )
        >>> pv_mag_div
        PV_MagDiv.kr()

    """

    _ordered_input_names = collections.OrderedDict(
        [("pv_chain_a", None), ("pv_chain_b", None), ("zeroed", 0.0001)]
    )


class PV_MagFreeze(PV_ChainUGen):
    """
    Freezes magnitudes.

    ::

        >>> pv_chain = supriya.ugens.FFT(
        ...     source=supriya.ugens.WhiteNoise.ar(),
        ... )
        >>> pv_mag_freeze = supriya.ugens.PV_MagFreeze.new(
        ...     pv_chain=pv_chain,
        ...     freeze=0,
        ... )
        >>> pv_mag_freeze
        PV_MagFreeze.kr()

    """

    _ordered_input_names = collections.OrderedDict([("pv_chain", None), ("freeze", 0)])


class PV_MagMul(PV_ChainUGen):
    """
    Multiplies FFT magnitudes.

    ::

        >>> pv_chain_a = supriya.ugens.FFT(
        ...     source=supriya.ugens.WhiteNoise.ar(),
        ... )
        >>> pv_chain_b = supriya.ugens.FFT(
        ...     source=supriya.ugens.LFSaw.ar(),
        ... )
        >>> pv_mag_mul = supriya.ugens.PV_MagMul.new(
        ...     pv_chain_a=pv_chain_a,
        ...     pv_chain_b=pv_chain_b,
        ... )
        >>> pv_mag_mul
        PV_MagMul.kr()

    """

    _ordered_input_names = collections.OrderedDict(
        [("pv_chain_a", None), ("pv_chain_b", None)]
    )


class PV_MagNoise(PV_ChainUGen):
    """
    Multiplies magnitudes by noise.

    ::

        >>> pv_chain = supriya.ugens.FFT(
        ...     source=supriya.ugens.WhiteNoise.ar(),
        ... )
        >>> pv_mag_noise = supriya.ugens.PV_MagNoise.new(
        ...     pv_chain=pv_chain,
        ... )
        >>> pv_mag_noise
        PV_MagNoise.kr()

    """

    _ordered_input_names = collections.OrderedDict([("pv_chain", None)])


class PV_MagShift(PV_ChainUGen):
    """
    Shifts and stretches magnitude bin position.

    ::

        >>> pv_chain = supriya.ugens.FFT(
        ...     source=supriya.ugens.WhiteNoise.ar(),
        ... )
        >>> pv_mag_shift = supriya.ugens.PV_MagShift.new(
        ...     pv_chain=pv_chain,
        ...     shift=0,
        ...     stretch=1,
        ... )
        >>> pv_mag_shift
        PV_MagShift.kr()

    """

    _ordered_input_names = collections.OrderedDict(
        [("pv_chain", None), ("stretch", 1), ("shift", 0)]
    )


class PV_MagSmear(PV_ChainUGen):
    """
    Averages magnitudes across bins.

    ::

        >>> pv_chain = supriya.ugens.FFT(
        ...     source=supriya.ugens.WhiteNoise.ar(),
        ... )
        >>> pv_mag_smear = supriya.ugens.PV_MagSmear.new(
        ...     bins=0,
        ...     pv_chain=pv_chain,
        ... )
        >>> pv_mag_smear
        PV_MagSmear.kr()

    """

    _ordered_input_names = collections.OrderedDict([("pv_chain", None), ("bins", 0)])


class PV_MagSquared(PV_ChainUGen):
    """
    Squares magnitudes.

    ::

        >>> pv_chain = supriya.ugens.FFT(
        ...     source=supriya.ugens.WhiteNoise.ar(),
        ... )
        >>> pv_mag_squared = supriya.ugens.PV_MagSquared.new(
        ...     pv_chain=pv_chain,
        ... )
        >>> pv_mag_squared
        PV_MagSquared.kr()

    """

    _ordered_input_names = collections.OrderedDict([("pv_chain", None)])


class PV_Max(PV_ChainUGen):
    """
    Maximum magnitude.

    ::

        >>> pv_chain_a = supriya.ugens.FFT(
        ...     source=supriya.ugens.WhiteNoise.ar(),
        ... )
        >>> pv_chain_b = supriya.ugens.FFT(
        ...     source=supriya.ugens.LFSaw.ar(),
        ... )
        >>> pv_max = supriya.ugens.PV_Max.new(
        ...     pv_chain_a=pv_chain_a,
        ...     pv_chain_b=pv_chain_b,
        ... )
        >>> pv_max
        PV_Max.kr()

    """

    _ordered_input_names = collections.OrderedDict(
        [("pv_chain_a", None), ("pv_chain_b", None)]
    )


class PV_Min(PV_ChainUGen):
    """
    Minimum magnitude.

    ::

        >>> pv_chain_a = supriya.ugens.FFT(
        ...     source=supriya.ugens.WhiteNoise.ar(),
        ... )
        >>> pv_chain_b = supriya.ugens.FFT(
        ...     source=supriya.ugens.LFSaw.ar(),
        ... )
        >>> pv_min = supriya.ugens.PV_Min.new(
        ...     pv_chain_a=pv_chain_a,
        ...     pv_chain_b=pv_chain_b,
        ... )
        >>> pv_min
        PV_Min.kr()

    """

    _ordered_input_names = collections.OrderedDict(
        [("pv_chain_a", None), ("pv_chain_b", None)]
    )


class PV_Mul(PV_ChainUGen):
    """
    Complex multiplication.

    ::

        >>> pv_chain_a = supriya.ugens.FFT(
        ...     source=supriya.ugens.WhiteNoise.ar(),
        ... )
        >>> pv_chain_b = supriya.ugens.FFT(
        ...     source=supriya.ugens.LFSaw.ar(),
        ... )
        >>> pv_mul = supriya.ugens.PV_Mul.new(
        ...     pv_chain_a=pv_chain_a,
        ...     pv_chain_b=pv_chain_b,
        ... )
        >>> pv_mul
        PV_Mul.kr()

    """

    _ordered_input_names = collections.OrderedDict(
        [("pv_chain_a", None), ("pv_chain_b", None)]
    )


class PV_PhaseShift(PV_ChainUGen):
    """
    Shifts phase.

    ::

        >>> pv_chain = supriya.ugens.FFT(
        ...     source=supriya.ugens.WhiteNoise.ar(),
        ... )
        >>> shift = supriya.ugens.LFNoise2.kr(1).scale(-1, 1, -180, 180)
        >>> pv_phase_shift = supriya.ugens.PV_PhaseShift.new(
        ...     pv_chain=pv_chain,
        ...     integrate=0,
        ...     shift=shift,
        ... )
        >>> pv_phase_shift
        PV_PhaseShift.kr()

    """

    _ordered_input_names = collections.OrderedDict(
        [("pv_chain", None), ("shift", None), ("integrate", 0)]
    )


class PV_PhaseShift270(PV_ChainUGen):
    """
    Shifts phase by 270 degrees.

    ::

        >>> pv_chain = supriya.ugens.FFT(
        ...     source=supriya.ugens.WhiteNoise.ar(),
        ... )
        >>> pv_phase_shift_270 = supriya.ugens.PV_PhaseShift270.new(
        ...     pv_chain=pv_chain,
        ... )
        >>> pv_phase_shift_270
        PV_PhaseShift270.kr()

    """

    _ordered_input_names = collections.OrderedDict([("pv_chain", None)])


class PV_PhaseShift90(PV_ChainUGen):
    """
    Shifts phase by 90 degrees.

    ::

        >>> pv_chain = supriya.ugens.FFT(
        ...     source=supriya.ugens.WhiteNoise.ar(),
        ... )
        >>> pv_phase_shift_90 = supriya.ugens.PV_PhaseShift90.new(
        ...     pv_chain=pv_chain,
        ... )
        >>> pv_phase_shift_90
        PV_PhaseShift90.kr()

    """

    _ordered_input_names = collections.OrderedDict([("pv_chain", None)])


class PV_RandComb(PV_ChainUGen):
    """
    Passes random bins.

    ::

        >>> pv_chain = supriya.ugens.FFT(
        ...     source=supriya.ugens.WhiteNoise.ar(),
        ... )
        >>> pv_rand_comb = supriya.ugens.PV_RandComb.new(
        ...     pv_chain=pv_chain,
        ...     trigger=0,
        ...     wipe=0,
        ... )
        >>> pv_rand_comb
        PV_RandComb.kr()

    """

    _ordered_input_names = collections.OrderedDict(
        [("pv_chain", None), ("wipe", 0), ("trigger", 0)]
    )


class PV_RandWipe(PV_ChainUGen):
    """
    Crossfades in random bin order.

    ::

        >>> pv_chain_a = supriya.ugens.FFT(
        ...     source=supriya.ugens.WhiteNoise.ar(),
        ... )
        >>> pv_chain_b = supriya.ugens.FFT(
        ...     source=supriya.ugens.LFSaw.ar(),
        ... )
        >>> pv_rand_wipe = supriya.ugens.PV_RandWipe.new(
        ...     pv_chain_a=pv_chain_a,
        ...     pv_chain_b=pv_chain_b,
        ...     trigger=0,
        ...     wipe=0,
        ... )
        >>> pv_rand_wipe
        PV_RandWipe.kr()

    """

    _ordered_input_names = collections.OrderedDict(
        [("pv_chain_a", None), ("pv_chain_b", None), ("wipe", 0), ("trigger", 0)]
    )


class PV_RectComb(PV_ChainUGen):
    """
    Makes gaps in the spectrum.

    ::

        >>> pv_chain = supriya.ugens.FFT(
        ...     source=supriya.ugens.WhiteNoise.ar(),
        ... )
        >>> pv_rect_comb = supriya.ugens.PV_RectComb.new(
        ...     pv_chain=pv_chain,
        ...     num_teeth=0,
        ...     phase=0,
        ...     width=0.5,
        ... )
        >>> pv_rect_comb
        PV_RectComb.kr()

    """

    _ordered_input_names = collections.OrderedDict(
        [("pv_chain", None), ("num_teeth", 0), ("phase", 0), ("width", 0.5)]
    )


class PV_RectComb2(PV_ChainUGen):
    """
    Makes gaps in the spectrum.

    ::

        >>> pv_chain_a = supriya.ugens.FFT(
        ...     source=supriya.ugens.WhiteNoise.ar(),
        ... )
        >>> pv_chain_b = supriya.ugens.FFT(
        ...     source=supriya.ugens.LFSaw.ar(),
        ... )
        >>> pv_rect_comb_2 = supriya.ugens.PV_RectComb2.new(
        ...     pv_chain_a=pv_chain_a,
        ...     pv_chain_b=pv_chain_b,
        ...     num_teeth=0,
        ...     phase=0,
        ...     width=0.5,
        ... )
        >>> pv_rect_comb_2
        PV_RectComb2.kr()

    """

    _ordered_input_names = collections.OrderedDict(
        [
            ("pv_chain_a", None),
            ("pv_chain_b", None),
            ("num_teeth", 0),
            ("phase", 0),
            ("width", 0.5),
        ]
    )


class RunningSum(UGen):
    """
    Tracks running sum over ``n`` frames.

    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> running_sum = supriya.ugens.RunningSum.ar(
        ...     sample_count=40,
        ...     source=source,
        ... )
        >>> running_sum
        RunningSum.ar()

    """

    _ordered_input_names = collections.OrderedDict(
        [("source", None), ("sample_count", 40)]
    )
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)



import collections
import struct
from collections.abc import Sequence

from supriya import CalculationRate, ParameterRate, utils
from supriya.system import SupriyaObject

from .bases import MultiOutUGen, UGen


class SynthDefCompiler(SupriyaObject):

    ### PUBLIC METHODS ###

    @staticmethod
    def compile_synthdef(synthdef, name):
        result = SynthDefCompiler.encode_string(name)
        result += synthdef._compiled_ugen_graph
        return result

    @staticmethod
    def compile_parameters(synthdef):
        result = []
        result.append(
            SynthDefCompiler.encode_unsigned_int_32bit(
                sum(len(_[1]) for _ in synthdef.indexed_parameters)
            )
        )
        for control_ugen in synthdef.control_ugens:
            for parameter in control_ugen.parameters:
                value = parameter.value
                if not isinstance(value, tuple):
                    value = (value,)
                for x in value:
                    result.append(SynthDefCompiler.encode_float(x))
        result.append(
            SynthDefCompiler.encode_unsigned_int_32bit(len(synthdef.indexed_parameters))
        )
        for index, parameter in synthdef.indexed_parameters:
            name = parameter.name
            result.append(SynthDefCompiler.encode_string(name))
            result.append(SynthDefCompiler.encode_unsigned_int_32bit(index))
        return bytes().join(result)

    @staticmethod
    def compile_synthdefs(synthdefs, use_anonymous_names=False):
        def flatten(value):
            if isinstance(value, Sequence) and not isinstance(
                value, (bytes, bytearray)
            ):
                return bytes().join(flatten(x) for x in value)
            return value

        result = []
        encoded_file_type_id = b"SCgf"
        result.append(encoded_file_type_id)
        encoded_file_version = SynthDefCompiler.encode_unsigned_int_32bit(2)
        result.append(encoded_file_version)
        encoded_synthdef_count = SynthDefCompiler.encode_unsigned_int_16bit(
            len(synthdefs)
        )
        result.append(encoded_synthdef_count)
        for synthdef in synthdefs:
            name = synthdef.name
            if not name or use_anonymous_names:
                name = synthdef.anonymous_name
            result.append(SynthDefCompiler.compile_synthdef(synthdef, name))
        result = flatten(result)
        result = bytes(result)
        return result

    @staticmethod
    def compile_ugen(ugen, synthdef):
        outputs = ugen._get_outputs()
        result = []
        result.append(SynthDefCompiler.encode_string(type(ugen).__name__))
        result.append(SynthDefCompiler.encode_unsigned_int_8bit(ugen.calculation_rate))
        result.append(SynthDefCompiler.encode_unsigned_int_32bit(len(ugen.inputs)))
        result.append(SynthDefCompiler.encode_unsigned_int_32bit(len(outputs)))
        result.append(
            SynthDefCompiler.encode_unsigned_int_16bit(int(ugen.special_index))
        )
        for input_ in ugen.inputs:
            result.append(SynthDefCompiler.compile_ugen_input_spec(input_, synthdef))
        for output in outputs:
            result.append(SynthDefCompiler.encode_unsigned_int_8bit(output))
        result = bytes().join(result)
        return result

    @staticmethod
    def compile_ugen_graph(synthdef):
        result = []
        result.append(
            SynthDefCompiler.encode_unsigned_int_32bit(len(synthdef.constants))
        )
        for constant in synthdef.constants:
            result.append(SynthDefCompiler.encode_float(constant))
        result.append(SynthDefCompiler.compile_parameters(synthdef))
        result.append(SynthDefCompiler.encode_unsigned_int_32bit(len(synthdef.ugens)))
        for ugen_index, ugen in enumerate(synthdef.ugens):
            result.append(SynthDefCompiler.compile_ugen(ugen, synthdef))
        result.append(SynthDefCompiler.encode_unsigned_int_16bit(0))
        result = bytes().join(result)
        return result

    @staticmethod
    def compile_ugen_input_spec(input_, synthdef):
        import supriya.synthdefs

        result = []
        if isinstance(input_, float):
            result.append(SynthDefCompiler.encode_unsigned_int_32bit(0xFFFFFFFF))
            constant_index = synthdef._constants.index(input_)
            result.append(SynthDefCompiler.encode_unsigned_int_32bit(constant_index))
        elif isinstance(input_, supriya.synthdefs.OutputProxy):
            ugen = input_.source
            output_index = input_.output_index
            ugen_index = synthdef._ugens.index(ugen)
            result.append(SynthDefCompiler.encode_unsigned_int_32bit(ugen_index))
            result.append(SynthDefCompiler.encode_unsigned_int_32bit(output_index))
        else:
            raise Exception("Unhandled input spec: {}".format(input_))
        return bytes().join(result)

    @staticmethod
    def encode_string(value):
        result = bytes(struct.pack(">B", len(value)))
        result += bytes(bytearray(value, encoding="ascii"))
        return result

    @staticmethod
    def encode_float(value):
        return bytes(struct.pack(">f", float(value)))

    @staticmethod
    def encode_unsigned_int_8bit(value):
        return bytes(struct.pack(">B", int(value)))

    @staticmethod
    def encode_unsigned_int_16bit(value):
        return bytes(struct.pack(">H", int(value)))

    @staticmethod
    def encode_unsigned_int_32bit(value):
        return bytes(struct.pack(">I", int(value)))


class SynthDefDecompiler(SupriyaObject):
    """
    SynthDef decompiler.

    ::

        >>> import supriya.synthdefs
        >>> import supriya.ugens
        >>> with supriya.synthdefs.SynthDefBuilder(
        ...     frequency=440,
        ...     trigger=supriya.synthdefs.Parameter(
        ...         value=0.0,
        ...         parameter_rate=supriya.ParameterRate.TRIGGER,
        ...     ),
        ... ) as builder:
        ...     sin_osc = supriya.ugens.SinOsc.ar(frequency=builder["frequency"])
        ...     decay = supriya.ugens.Decay.kr(
        ...         decay_time=0.5,
        ...         source=builder["trigger"],
        ...     )
        ...     enveloped_sin = sin_osc * decay
        ...     out = supriya.ugens.Out.ar(bus=0, source=enveloped_sin)
        ...
        >>> synthdef = builder.build()
        >>> supriya.graph(synthdef)  # doctest: +SKIP

    ::

        >>> print(synthdef)
        synthdef:
            name: 001520731aee5371fefab6b505cf64dd
            ugens:
            -   TrigControl.kr: null
            -   Decay.kr:
                    decay_time: 0.5
                    source: TrigControl.kr[0:trigger]
            -   Control.kr: null
            -   SinOsc.ar:
                    frequency: Control.kr[0:frequency]
                    phase: 0.0
            -   BinaryOpUGen(MULTIPLICATION).ar:
                    left: SinOsc.ar[0]
                    right: Decay.kr[0]
            -   Out.ar:
                    bus: 0.0
                    source[0]: BinaryOpUGen(MULTIPLICATION).ar[0]

    ::

        >>> compiled_synthdef = synthdef.compile()
        >>> sdd = supriya.synthdefs.SynthDefDecompiler
        >>> decompiled_synthdef = sdd.decompile_synthdefs(compiled_synthdef)[0]
        >>> supriya.graph(decompiled_synthdef)  # doctest: +SKIP

    ::

        >>> print(decompiled_synthdef)
        synthdef:
            name: 001520731aee5371fefab6b505cf64dd
            ugens:
            -   TrigControl.kr: null
            -   Decay.kr:
                    decay_time: 0.5
                    source: TrigControl.kr[0:trigger]
            -   Control.kr: null
            -   SinOsc.ar:
                    frequency: Control.kr[0:frequency]
                    phase: 0.0
            -   BinaryOpUGen(MULTIPLICATION).ar:
                    left: SinOsc.ar[0]
                    right: Decay.kr[0]
            -   Out.ar:
                    bus: 0.0
                    source[0]: BinaryOpUGen(MULTIPLICATION).ar[0]

    ::

        >>> str(synthdef) == str(decompiled_synthdef)
        True

    """

    ### PRIVATE METHODS ###

    @staticmethod
    def _decode_constants(value, index):
        sdd = SynthDefDecompiler
        constants = []
        constants_count, index = sdd._decode_int_32bit(value, index)
        for _ in range(constants_count):
            constant, index = sdd._decode_float(value, index)
            constants.append(constant)
        return constants, index

    @staticmethod
    def _decode_parameters(value, index):
        import supriya.synthdefs

        sdd = SynthDefDecompiler
        parameter_values = []
        parameter_count, index = sdd._decode_int_32bit(value, index)
        for _ in range(parameter_count):
            parameter_value, index = sdd._decode_float(value, index)
            parameter_values.append(parameter_value)
        parameter_count, index = sdd._decode_int_32bit(value, index)
        parameter_names = []
        parameter_indices = []
        for _ in range(parameter_count):
            parameter_name, index = sdd._decode_string(value, index)
            parameter_index, index = sdd._decode_int_32bit(value, index)
            parameter_names.append(parameter_name)
            parameter_indices.append(parameter_index)
        indexed_parameters = []
        if parameter_count:
            pairs = tuple(zip(parameter_indices, parameter_names))
            pairs = sorted(pairs, key=lambda x: x[0])
            iterator = utils.iterate_nwise(pairs)
            for (index_one, name_one), (index_two, name_two) in iterator:
                value = parameter_values[index_one:index_two]
                if len(value) == 1:
                    value = value[0]
                parameter = supriya.synthdefs.Parameter(name=name_one, value=value)
                indexed_parameters.append((index_one, parameter))
            index_one, name_one = pairs[-1]
            value = parameter_values[index_one:]
            if len(value) == 1:
                value = value[0]
            parameter = supriya.synthdefs.Parameter(name=name_one, value=value)
            indexed_parameters.append((index_one, parameter))
            indexed_parameters.sort(key=lambda x: parameter_names.index(x[1].name))
        indexed_parameters = collections.OrderedDict(indexed_parameters)
        return indexed_parameters, index

    @staticmethod
    def _decompile_synthdef(value, index):
        import supriya.synthdefs
        import supriya.ugens

        sdd = SynthDefDecompiler
        synthdef = None
        name, index = sdd._decode_string(value, index)
        constants, index = sdd._decode_constants(value, index)
        indexed_parameters, index = sdd._decode_parameters(value, index)
        ugens = []
        ugen_count, index = sdd._decode_int_32bit(value, index)
        for i in range(ugen_count):
            ugen_name, index = sdd._decode_string(value, index)
            calculation_rate, index = sdd._decode_int_8bit(value, index)
            calculation_rate = supriya.CalculationRate(calculation_rate)
            input_count, index = sdd._decode_int_32bit(value, index)
            output_count, index = sdd._decode_int_32bit(value, index)
            special_index, index = sdd._decode_int_16bit(value, index)
            inputs = []
            for _ in range(input_count):
                ugen_index, index = sdd._decode_int_32bit(value, index)
                if ugen_index == 0xFFFFFFFF:
                    constant_index, index = sdd._decode_int_32bit(value, index)
                    constant_index = int(constant_index)
                    inputs.append(constants[constant_index])
                else:
                    ugen = ugens[ugen_index]
                    ugen_output_index, index = sdd._decode_int_32bit(value, index)
                    output_proxy = ugen[ugen_output_index]
                    inputs.append(output_proxy)
            for _ in range(output_count):
                output_rate, index = sdd._decode_int_8bit(value, index)
            ugen_class = getattr(supriya.ugens, ugen_name, None)
            if ugen_class is None:
                ugen_class = getattr(supriya.synthdefs, ugen_name)
            ugen = supriya.synthdefs.UGen.__new__(ugen_class)
            if issubclass(ugen_class, supriya.synthdefs.Control):
                starting_control_index = special_index
                parameters = sdd._collect_parameters_for_control(
                    calculation_rate,
                    indexed_parameters,
                    inputs,
                    output_count,
                    starting_control_index,
                    ugen_class,
                )
                ugen_class.__init__(
                    ugen,
                    parameters=parameters,
                    starting_control_index=starting_control_index,
                    calculation_rate=calculation_rate,
                )
            else:
                kwargs = {}
                if not ugen._unexpanded_input_names:
                    for i, input_name in enumerate(ugen._ordered_input_names):
                        kwargs[input_name] = inputs[i]
                else:
                    for i, input_name in enumerate(ugen._ordered_input_names):
                        if input_name not in ugen._unexpanded_input_names:
                            kwargs[input_name] = inputs[i]
                        else:
                            kwargs[input_name] = tuple(inputs[i:])
                if issubclass(ugen_class, MultiOutUGen):
                    MultiOutUGen.__init__(
                        ugen,
                        calculation_rate=calculation_rate,
                        channel_count=output_count,
                        special_index=special_index,
                        **kwargs,
                    )
                else:
                    UGen.__init__(
                        ugen,
                        calculation_rate=calculation_rate,
                        special_index=special_index,
                        **kwargs,
                    )
            ugens.append(ugen)
        variants_count, index = sdd._decode_int_16bit(value, index)
        synthdef = supriya.synthdefs.SynthDef(ugens=ugens, name=name, decompiled=True)
        if synthdef.name == synthdef.anonymous_name:
            synthdef._name = None
        return synthdef, index

    @staticmethod
    def _decode_string(value, index):
        length = struct.unpack(">B", value[index : index + 1])[0]
        index += 1
        result = value[index : index + length]
        result = result.decode("ascii")
        index += length
        return result, index

    @staticmethod
    def _decode_float(value, index):
        result = struct.unpack(">f", value[index : index + 4])[0]
        index += 4
        return result, index

    @staticmethod
    def _decode_int_8bit(value, index):
        result = struct.unpack(">B", value[index : index + 1])[0]
        index += 1
        return result, index

    @staticmethod
    def _decode_int_16bit(value, index):
        result = struct.unpack(">H", value[index : index + 2])[0]
        index += 2
        return result, index

    @staticmethod
    def _decode_int_32bit(value, index):
        result = struct.unpack(">I", value[index : index + 4])[0]
        index += 4
        return result, index

    @staticmethod
    def _collect_parameters_for_control(
        calculation_rate,
        indexed_parameters,
        inputs,
        output_count,
        starting_control_index,
        ugen_class,
    ):
        import supriya.synthdefs
        import supriya.ugens

        parameter_rate = ParameterRate.CONTROL
        if issubclass(ugen_class, supriya.synthdefs.TrigControl):
            parameter_rate = ParameterRate.TRIGGER
        elif calculation_rate == CalculationRate.SCALAR:
            parameter_rate = ParameterRate.SCALAR
        elif calculation_rate == CalculationRate.AUDIO:
            parameter_rate = ParameterRate.AUDIO
        parameters = []
        collected_output_count = 0
        lag = 0.0
        while collected_output_count < output_count:
            if inputs:
                lag = inputs[collected_output_count]
            parameter = indexed_parameters[
                starting_control_index + collected_output_count
            ]
            parameter._parameter_rate = parameter_rate
            if lag:
                parameter._lag = lag
            parameters.append(parameter)
            collected_output_count += len(parameter)
        return parameters

    ### PUBLIC METHODS ###

    @staticmethod
    def decompile_synthdef(value):
        synthdefs = SynthDefDecompiler.decompile_synthdefs(value)
        assert len(synthdefs) == 1
        return synthdefs[0]

    @staticmethod
    def decompile_synthdefs(value):
        synthdefs = []
        sdd = SynthDefDecompiler
        index = 4
        assert value[:index] == b"SCgf"
        file_version, index = sdd._decode_int_32bit(value, index)
        synthdef_count, index = sdd._decode_int_16bit(value, index)
        for _ in range(synthdef_count):
            synthdef, index = sdd._decompile_synthdef(value, index)
            synthdefs.append(synthdef)
        return synthdefs



import collections

from supriya import CalculationRate
from supriya.synthdefs import UGen


class CheckBadValues(UGen):
    """
    Tests for infinity, not-a-number, and denormals.

    ::

        >>> source = supriya.ugens.SinOsc.ar()
        >>> ugen_id = 23
        >>> post_mode = 0
        >>> check_bad_values = supriya.ugens.CheckBadValues.ar(
        ...     source=source,
        ...     ugen_id=ugen_id,
        ...     post_mode=post_mode,
        ... )
        >>> check_bad_values
        CheckBadValues.ar()

    """

    ### CLASS VARIABLES ###

    _ordered_input_names = collections.OrderedDict(
        [("source", None), ("ugen_id", 0.0), ("post_mode", 2.0)]
    )
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)

    ### INITIALIZER ###

    def __init__(self, calculation_rate=None, ugen_id=0, post_mode=2, source=None):
        assert int(post_mode) in (0, 1, 2)
        UGen.__init__(
            self,
            calculation_rate=calculation_rate,
            ugen_id=ugen_id,
            post_mode=post_mode,
            source=source,
        )


class Sanitize(UGen):
    """
    Remove infinity, NaN, and denormals.
    """

    _ordered_input_names = collections.OrderedDict([("source", None), ("replace", 0.0)])
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)



import collections

from supriya import CalculationRate
from supriya.synthdefs import PseudoUGen, UGen

from .delay import DelayN


class Amplitude(UGen):
    """
    An amplitude follower.

    ::

        >>> source = supriya.ugens.In.ar(0)
        >>> amplitude = supriya.ugens.Amplitude.kr(
        ...     attack_time=0.01,
        ...     release_time=0.01,
        ...     source=source,
        ... )
        >>> amplitude
        Amplitude.kr()

    """

    _ordered_input_names = collections.OrderedDict(
        [("source", None), ("attack_time", 0.01), ("release_time", 0.01)]
    )
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)


class Compander(UGen):
    """
    A general purpose hard-knee dynamics processor.

    """

    _ordered_input_names = collections.OrderedDict(
        [
            ("source", None),
            ("control", 0.0),
            ("threshold", 0.5),
            ("slope_below", 1.0),
            ("slope_above", 1.0),
            ("clamp_time", 0.01),
            ("relax_time", 0.1),
        ]
    )
    _valid_calculation_rates = (CalculationRate.AUDIO,)


class CompanderD(PseudoUGen):
    """
    A convenience constructor for Compander.
    """

    ### PUBLIC METHODS ###

    @classmethod
    def ar(
        cls,
        source=None,
        threshold=0.5,
        clamp_time=0.01,
        relax_time=0.1,
        slope_above=1.0,
        slope_below=1.0,
    ):
        """
        Constructs an audio-rate dynamics processor.

        ..  container:: example

            ::

                >>> source = supriya.ugens.In.ar(bus=0)
                >>> compander_d = supriya.ugens.CompanderD.ar(
                ...     source=source,
                ... )
                >>> supriya.graph(compander_d)  # doctest: +SKIP

            ::

                >>> print(compander_d)
                synthdef:
                    name: d4e7b88df56af5070a88f09b0f8c633e
                    ugens:
                    -   In.ar:
                            bus: 0.0
                    -   DelayN.ar:
                            delay_time: 0.01
                            maximum_delay_time: 0.01
                            source: In.ar[0]
                    -   Compander.ar:
                            clamp_time: 0.01
                            control: DelayN.ar[0]
                            relax_time: 0.1
                            slope_above: 1.0
                            slope_below: 1.0
                            source: In.ar[0]
                            threshold: 0.5

        Returns ugen graph.
        """
        control = DelayN.ar(
            source=source, maximum_delay_time=clamp_time, delay_time=clamp_time
        )
        return Compander._new_expanded(
            clamp_time=clamp_time,
            calculation_rate=CalculationRate.AUDIO,
            relax_time=relax_time,
            slope_above=slope_above,
            slope_below=slope_below,
            source=source,
            control=control,
            threshold=threshold,
        )


class Limiter(UGen):
    """
    A peak limiter.

    ::

        >>> source = supriya.ugens.In.ar(0)
        >>> limiter = supriya.ugens.Limiter.ar(
        ...     duration=0.01,
        ...     level=1,
        ...     source=source,
        ... )
        >>> limiter
        Limiter.ar()

    """

    _ordered_input_names = collections.OrderedDict(
        [("source", None), ("level", 1), ("duration", 0.01)]
    )
    _valid_calculation_rates = (CalculationRate.AUDIO,)


class Normalizer(UGen):
    """
    A dynamics flattener.

    ::

        >>> source = supriya.ugens.In.ar(0)
        >>> normalizer = supriya.ugens.Normalizer.ar(
        ...     duration=0.01,
        ...     level=1,
        ...     source=source,
        ... )
        >>> normalizer
        Normalizer.ar()

    """

    _ordered_input_names = collections.OrderedDict(
        [("source", None), ("level", 1), ("duration", 0.01)]
    )
    _valid_calculation_rates = (CalculationRate.AUDIO,)



import contextlib
import socket

from supriya import utils


def find_free_port():
    with contextlib.closing(socket.socket(socket.AF_INET, socket.SOCK_DGRAM)) as s:
        s.bind(("", 0))
        s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        return s.getsockname()[1]


def format_datagram(datagram):
    result = []
    result.append("size {}".format(len(datagram)))
    index = 0
    while index < len(datagram):
        chunk = datagram[index : index + 16]
        line = "{: >4}   ".format(index)
        hex_blocks = []
        ascii_block = ""
        for chunk in utils.group_iterable_by_count(chunk, 4):
            hex_block = []
            for byte in chunk:
                char = int(byte)
                if 31 < char < 127:
                    char = chr(char)
                else:
                    char = "."
                ascii_block += char
                hexed = hex(byte)[2:].zfill(2)
                hex_block.append(hexed)
            hex_block = " ".join(hex_block)
            hex_blocks.append(hex_block)
        hex_blocks = "  ".join(hex_blocks)
        ascii_block = "|{}|".format(ascii_block)
        hex_blocks = "{: <53}".format(hex_blocks)
        line += hex_blocks
        line += ascii_block
        result.append(line)
        index += 16
    result = "\n".join(result)
    return result



import collections

from supriya import CalculationRate
from supriya.synthdefs import UGen
from supriya.typing import UGenInputMap


class BlockSize(UGen):
    """
    A block size info unit generator.

    ::

        >>> supriya.ugens.BlockSize.ir()
        BlockSize.ir()

    """

    _ordered_input_names: UGenInputMap = collections.OrderedDict([])
    _valid_calculation_rates = (CalculationRate.SCALAR,)


class BufChannels(UGen):
    """
    A buffer channel count info unit generator.

    ::

        >>> supriya.ugens.BufChannels.kr(buffer_id=0)
        BufChannels.kr()

    """

    _ordered_input_names = collections.OrderedDict([("buffer_id", None)])
    _valid_calculation_rates = (CalculationRate.CONTROL, CalculationRate.SCALAR)


class BufDur(UGen):
    """
    A buffer duration info unit generator.

    ::

        >>> supriya.ugens.BufDur.kr(buffer_id=0)
        BufDur.kr()

    """

    _ordered_input_names = collections.OrderedDict([("buffer_id", None)])
    _valid_calculation_rates = (CalculationRate.CONTROL, CalculationRate.SCALAR)


class BufFrames(UGen):
    """
    A buffer frame count info unit generator.

    ::

        >>> supriya.ugens.BufFrames.kr(buffer_id=0)
        BufFrames.kr()

    """

    _ordered_input_names = collections.OrderedDict([("buffer_id", None)])
    _valid_calculation_rates = (CalculationRate.CONTROL, CalculationRate.SCALAR)


class BufRateScale(UGen):
    """
    A buffer sample-rate scale info unit generator.

    ::

        >>> supriya.ugens.BufRateScale.kr(buffer_id=0)
        BufRateScale.kr()

    """

    _ordered_input_names = collections.OrderedDict([("buffer_id", None)])
    _valid_calculation_rates = (CalculationRate.CONTROL, CalculationRate.SCALAR)


class BufSampleRate(UGen):
    """
    A buffer sample-rate info unit generator.

    ::

        >>> supriya.ugens.BufSampleRate.kr(buffer_id=0)
        BufSampleRate.kr()

    """

    _ordered_input_names = collections.OrderedDict([("buffer_id", None)])
    _valid_calculation_rates = (CalculationRate.CONTROL, CalculationRate.SCALAR)


class BufSamples(UGen):
    """
    A buffer sample count info unit generator.

    ::

        >>> supriya.ugens.BufSamples.kr(buffer_id=0)
        BufSamples.kr()

    """

    _ordered_input_names = collections.OrderedDict([("buffer_id", None)])
    _valid_calculation_rates = (CalculationRate.CONTROL, CalculationRate.SCALAR)


class ControlDur(UGen):
    """
    A control duration info unit generator.

    ::

        >>> supriya.ugens.ControlDur.ir()
        ControlDur.ir()

    """

    _ordered_input_names: UGenInputMap = collections.OrderedDict([])
    _valid_calculation_rates = (CalculationRate.SCALAR,)


class ControlRate(UGen):
    """
    A control rate info unit generator.

    ::

        >>> supriya.ugens.ControlRate.ir()
        ControlRate.ir()

    """

    _ordered_input_names: UGenInputMap = collections.OrderedDict([])
    _valid_calculation_rates = (CalculationRate.SCALAR,)


class NodeID(UGen):
    """
    A node ID info unit generator.

    ::

        >>> supriya.ugens.NodeID.ir()
        NodeID.ir()

    """

    _ordered_input_names: UGenInputMap = collections.OrderedDict([])
    _valid_calculation_rates = (CalculationRate.SCALAR,)


class NumAudioBuses(UGen):
    """
    A number of audio buses info unit generator.

    ::

        >>> supriya.ugens.NumAudioBuses.ir()
        NumAudioBuses.ir()

    """

    _ordered_input_names: UGenInputMap = collections.OrderedDict([])
    _valid_calculation_rates = (CalculationRate.SCALAR,)


class NumBuffers(UGen):
    """
    A number of buffers info unit generator.

    ::

        >>> supriya.ugens.NumBuffers.ir()
        NumBuffers.ir()

    """

    _ordered_input_names: UGenInputMap = collections.OrderedDict([])
    _valid_calculation_rates = (CalculationRate.SCALAR,)


class NumControlBuses(UGen):
    """
    A number of control buses info unit generator.

    ::

        >>> supriya.ugens.NumControlBuses.ir()
        NumControlBuses.ir()

    """

    _ordered_input_names: UGenInputMap = collections.OrderedDict([])
    _valid_calculation_rates = (CalculationRate.SCALAR,)


class NumInputBuses(UGen):
    """
    A number of input buses info unit generator.

    ::

        >>> supriya.ugens.NumInputBuses.ir()
        NumInputBuses.ir()

    """

    _ordered_input_names: UGenInputMap = collections.OrderedDict([])
    _valid_calculation_rates = (CalculationRate.SCALAR,)


class NumOutputBuses(UGen):
    """
    A number of output buses info unit generator.

    ::

        >>> supriya.ugens.NumOutputBuses.ir()
        NumOutputBuses.ir()

    """

    _ordered_input_names: UGenInputMap = collections.OrderedDict([])
    _valid_calculation_rates = (CalculationRate.SCALAR,)


class NumRunningSynths(UGen):
    """
    A number of running synths info unit generator.

    ::

        >>> supriya.ugens.NumRunningSynths.ir()
        NumRunningSynths.ir()

    """

    _ordered_input_names: UGenInputMap = collections.OrderedDict([])
    _valid_calculation_rates = (CalculationRate.CONTROL, CalculationRate.SCALAR)


class RadiansPerSample(UGen):
    """
    A radians-per-sample info unit generator.

    ::

        >>> supriya.ugens.RadiansPerSample.ir()
        RadiansPerSample.ir()

    """

    _ordered_input_names: UGenInputMap = collections.OrderedDict([])
    _valid_calculation_rates = (CalculationRate.SCALAR,)


class SampleDur(UGen):
    """
    A sample duration info unit generator.

    ::

        >>> supriya.ugens.SampleDur.ir()
        SampleDur.ir()

    """

    _ordered_input_names: UGenInputMap = collections.OrderedDict([])
    _valid_calculation_rates = (CalculationRate.SCALAR,)


class SampleRate(UGen):
    """
    A sample-rate info unit generator.

    ::

        >>> supriya.ugens.SampleRate.ir()
        SampleRate.ir()

    """

    _ordered_input_names: UGenInputMap = collections.OrderedDict([])
    _valid_calculation_rates = (CalculationRate.SCALAR,)


class SubsampleOffset(UGen):
    """
    A subsample-offset info unit generator.

    ::

        >>> supriya.ugens.SubsampleOffset.ir()
        SubsampleOffset.ir()

    """

    _ordered_input_names: UGenInputMap = collections.OrderedDict([])
    _valid_calculation_rates = (CalculationRate.SCALAR,)



import collections

from supriya import CalculationRate, SignalRange
from supriya.synthdefs import UGen


class KeyState(UGen):
    _ordered_input_names = collections.OrderedDict(
        [("keycode", 0.0), ("minimum", 0.0), ("maximum", 1.0), ("lag", 0.2)]
    )
    _signal_range = SignalRange.UNIPOLAR
    _valid_calculation_rates = (CalculationRate.CONTROL,)


class MouseButton(UGen):
    """
    A mouse-button tracker.

    ::

        >>> supriya.ugens.MouseButton.kr()
        MouseButton.kr()

    """

    _ordered_input_names = collections.OrderedDict(
        [("minimum", 0), ("maximum", 1), ("lag", 0.2)]
    )
    _signal_range = SignalRange.UNIPOLAR
    _valid_calculation_rates = (CalculationRate.CONTROL,)


class MouseX(UGen):
    """
    A mouse cursor tracker.

    MouseX tracks the y-axis of the mouse cursor position.

    ::

        >>> supriya.ugens.MouseX.kr()
        MouseX.kr()

    """

    _ordered_input_names = collections.OrderedDict(
        [("minimum", 0), ("maximum", 1), ("warp", 0), ("lag", 0.2)]
    )
    _signal_range = SignalRange.UNIPOLAR
    _valid_calculation_rates = (CalculationRate.CONTROL,)


class MouseY(UGen):
    """
    A mouse cursor tracker.

    MouseY tracks the y-axis of the mouse cursor position.

    ::

        >>> supriya.ugens.MouseY.kr()
        MouseY.kr()

    """

    _ordered_input_names = collections.OrderedDict(
        [("minimum", 0), ("maximum", 1), ("warp", 0), ("lag", 0.2)]
    )
    _signal_range = SignalRange.UNIPOLAR
    _valid_calculation_rates = (CalculationRate.CONTROL,)



import asyncio
import atexit
import dataclasses
import logging
import queue
import socketserver
import threading
import time
from collections.abc import Sequence
from typing import Any, Callable, Dict, NamedTuple, Optional, Set, Tuple, Union

from .captures import Capture, CaptureEntry
from .messages import OscBundle, OscMessage

osc_protocol_logger = logging.getLogger("supriya.osc.protocol")
osc_in_logger = logging.getLogger("supriya.osc.in")
osc_out_logger = logging.getLogger("supriya.osc.out")
udp_in_logger = logging.getLogger("supriya.udp.in")
udp_out_logger = logging.getLogger("supriya.udp.out")


class OscProtocolOffline(Exception):
    pass


class OscProtocolAlreadyConnected(Exception):
    pass


class OscCallback(NamedTuple):
    pattern: Tuple[Union[str, int, float], ...]
    procedure: Callable
    failure_pattern: Optional[Tuple[Union[str, int, float], ...]] = None
    once: bool = False


@dataclasses.dataclass
class HealthCheck:
    request_pattern: str
    response_pattern: str
    callback: Callable
    timeout: float = 1.0
    backoff_factor: float = 1.5
    max_attempts: int = 5


class OscProtocol:

    ### INITIALIZER ###

    def __init__(self):
        self.callbacks: Dict[Any, Any] = {}
        self.captures: Set[Capture] = set()
        self.healthcheck = None
        self.healthcheck_osc_callback = None
        self.attempts = 0
        self.ip_address = None
        self.is_running: bool = False
        self.port = None

    ### PRIVATE METHODS ###

    def _add_callback(self, callback: OscCallback):
        patterns = [callback.pattern]
        if callback.failure_pattern:
            patterns.append(callback.failure_pattern)
        for pattern in patterns:
            callback_map = self.callbacks
            for item in pattern:
                callbacks, callback_map = callback_map.setdefault(item, ([], {}))
            callbacks.append(callback)

    def _match_callbacks(self, message):
        items = (message.address,) + message.contents
        matching_callbacks = []
        callback_map = self.callbacks
        for item in items:
            if item not in callback_map:
                break
            callbacks, callback_map = callback_map[item]
            matching_callbacks.extend(callbacks)
        for callback in matching_callbacks:
            if callback.once:
                self.unregister(callback)
        return matching_callbacks

    def _remove_callback(self, callback: OscCallback):
        def delete(pattern, original_callback_map):
            key = pattern.pop(0)
            if key not in original_callback_map:
                return
            callbacks, callback_map = original_callback_map[key]
            if pattern:
                delete(pattern, callback_map)
            if callback in callbacks:
                callbacks.remove(callback)
            if not callbacks and not callback_map:
                original_callback_map.pop(key)

        patterns = [callback.pattern]
        if callback.failure_pattern:
            patterns.append(callback.failure_pattern)
        for pattern in patterns:
            delete(list(pattern), self.callbacks)

    def _pass_healthcheck(self, message):
        osc_protocol_logger.info(f"{self.ip_address}:{self.port} ...healthcheck passed")
        self.attempts = 0

    def _setup(self, ip_address, port, healthcheck):
        self.ip_address = ip_address
        self.port = port
        self.healthcheck = healthcheck
        if self.healthcheck:
            self.healthcheck_osc_callback = self.register(
                pattern=self.healthcheck.response_pattern,
                procedure=self._pass_healthcheck,
            )

    def _teardown(self):
        self.is_running = False
        if self.healthcheck is not None:
            self.unregister(self.healthcheck_osc_callback)

    def _validate_callback(
        self, pattern, procedure, *, failure_pattern=None, once=False
    ):
        if isinstance(pattern, (str, int, float)):
            pattern = [pattern]
        if isinstance(failure_pattern, (str, int, float)):
            failure_pattern = [failure_pattern]
        assert callable(procedure)
        return OscCallback(
            pattern=tuple(pattern),
            failure_pattern=failure_pattern,
            procedure=procedure,
            once=bool(once),
        )

    def _validate_receive(self, datagram):
        udp_in_logger.debug(f"{self.ip_address}:{self.port} {datagram}")
        try:
            message = OscMessage.from_datagram(datagram)
        except Exception:
            raise
        osc_in_logger.debug(f"{self.ip_address}:{self.port} {message!r}")
        for callback in self._match_callbacks(message):
            callback.procedure(message)
        for capture in self.captures:
            capture.messages.append(
                CaptureEntry(timestamp=time.time(), label="R", message=message)
            )

    def _validate_send(self, message):
        if not self.is_running:
            raise OscProtocolOffline
        if not isinstance(message, (str, Sequence, OscBundle, OscMessage)):
            raise ValueError(message)
        if isinstance(message, str):
            message = OscMessage(message)
        elif isinstance(message, Sequence):
            message = OscMessage(*message)
        osc_out_logger.debug(f"{self.ip_address}:{self.port} {message!r}")
        for capture in self.captures:
            capture.messages.append(
                CaptureEntry(timestamp=time.time(), label="S", message=message)
            )
        datagram = message.to_datagram()
        udp_out_logger.debug(f"{self.ip_address}:{self.port} {datagram}")
        return datagram

    ### PUBLIC METHODS ###

    def capture(self):
        return Capture(self)

    def connect(self, ip_address: str, port: int, *, healthcheck: HealthCheck = None):
        ...

    def disconnect(self):
        ...

    def register(
        self, pattern, procedure, *, failure_pattern=None, once=False
    ) -> OscCallback:
        ...

    def send(self, message):
        ...

    def unregister(self, callback: OscCallback):
        ...


class AsyncOscProtocol(asyncio.DatagramProtocol, OscProtocol):

    ### INITIALIZER ###

    def __init__(self):
        asyncio.DatagramProtocol.__init__(self)
        OscProtocol.__init__(self)
        self.loop = None

    ### PRIVATE METHODS ###

    async def _run_healthcheck(self):
        while self.is_running:
            sleep_time = self.healthcheck.timeout * pow(
                self.healthcheck.backoff_factor, self.attempts
            )
            self.attempts += 1
            if self.attempts >= self.healthcheck.max_attempts:
                self.exit_future.set_result(True)
                self._teardown()
                self.transport.close()
                obj_ = self.healthcheck.callback()
                if asyncio.iscoroutine(obj_):
                    self.loop.create_task(obj_)
                return
            self.send(OscMessage(*self.healthcheck.request_pattern))
            await asyncio.sleep(sleep_time)

    ### PUBLIC METHODS ###

    async def connect(
        self, ip_address: str, port: int, *, healthcheck: HealthCheck = None
    ):
        if self.is_running:
            raise OscProtocolAlreadyConnected
        self._setup(ip_address, port, healthcheck)
        self.loop = asyncio.get_running_loop()
        self.exit_future = self.loop.create_future()
        _, protocol = await self.loop.create_datagram_endpoint(
            lambda: self, remote_addr=(ip_address, port)
        )

    def connection_made(self, transport):
        loop = asyncio.get_running_loop()
        self.transport = transport
        self.is_running = True
        if self.healthcheck:
            self.healthcheck_task = loop.create_task(self._run_healthcheck())

    def connection_lost(self, exc):
        pass

    def datagram_received(self, data, addr):
        self._validate_receive(data)

    async def disconnect(self):
        if not self.is_running:
            return
        self.exit_future.set_result(True)
        self._teardown()
        if self.loop.is_closed():
            return
        if not self.transport.is_closing():
            self.transport.close()
        if self.healthcheck is not None:
            await self.healthcheck_task

    def error_received(self, exc):
        osc_out_logger.warning(f"{self.ip_address}:{self.port} {exc}")

    def register(
        self, pattern, procedure, *, failure_pattern=None, once=False
    ) -> OscCallback:
        callback = self._validate_callback(
            pattern, procedure, failure_pattern=failure_pattern, once=once
        )
        self._add_callback(callback)
        return callback

    def send(self, message):
        datagram = self._validate_send(message)
        return self.transport.sendto(datagram)

    def unregister(self, callback: OscCallback):
        self._remove_callback(callback)


class ThreadedOscServer(socketserver.UDPServer):
    osc_protocol: "ThreadedOscProtocol"

    def verify_request(self, request, client_address):
        self.osc_protocol._process_command_queue()
        return True

    def service_actions(self):
        self.osc_protocol._run_healthcheck()


class ThreadedOscHandler(socketserver.BaseRequestHandler):
    def handle(self):
        data = self.request[0]
        self.server.osc_protocol._validate_receive(data)


class ThreadedOscProtocol(OscProtocol):

    ### INITIALIZER ###

    def __init__(self):
        OscProtocol.__init__(self)
        self.command_queue = queue.Queue()
        self.lock = threading.RLock()
        self.osc_server = None
        self.osc_server_thread = None
        atexit.register(self.disconnect)

    ### PRIVATE METHODS ###

    def _process_command_queue(self):
        while self.command_queue.qsize():
            try:
                action, callback = self.command_queue.get()
            except queue.Empty:
                continue
            if action == "add":
                self._add_callback(callback)
            elif action == "remove":
                self._remove_callback(callback)

    def _run_healthcheck(self):
        if self.healthcheck is None:
            return
        now = time.time()
        if now < self.healthcheck_deadline:
            return
        if self.attempts > 0:
            remaining = self.healthcheck.max_attempts - self.attempts
            osc_protocol_logger.info(
                f"{self.ip_address}:{self.port} healthcheck failed, {remaining} attempts remaining"
            )
        new_timeout = self.healthcheck.timeout * pow(
            self.healthcheck.backoff_factor, self.attempts
        )
        self.healthcheck_deadline = now + new_timeout
        self.attempts += 1
        if self.attempts <= self.healthcheck.max_attempts:
            osc_protocol_logger.info(f"{self.ip_address}:{self.port} healthchecking...")
            self.send(OscMessage(*self.healthcheck.request_pattern))
            return
        osc_protocol_logger.info(
            f"{self.ip_address}:{self.port} healthcheck failure limit exceeded"
        )
        self.osc_server._BaseServer__shutdown_request = True
        self.disconnect()
        self.healthcheck.callback()

    def _server_factory(self, ip_address, port):
        server = ThreadedOscServer(
            (self.ip_address, self.port), ThreadedOscHandler, bind_and_activate=False
        )
        server.osc_protocol = self
        return server

    ### PUBLIC METHODS ###

    def connect(self, ip_address: str, port: int, *, healthcheck: HealthCheck = None):
        osc_protocol_logger.info(f"{self.ip_address}:{self.port} connecting...")
        if self.is_running:
            osc_protocol_logger.info(
                f"{self.ip_address}:{self.port} already connected!"
            )
            raise OscProtocolAlreadyConnected
        self._setup(ip_address, port, healthcheck)
        self.healthcheck_deadline = time.time()
        self.osc_server = self._server_factory(ip_address, port)
        self.osc_server_thread = threading.Thread(target=self.osc_server.serve_forever)
        self.osc_server_thread.daemon = True
        self.osc_server_thread.start()
        self.is_running = True
        osc_protocol_logger.info(f"{self.ip_address}:{self.port} ...connected")

    def disconnect(self):
        osc_protocol_logger.info(f"{self.ip_address}:{self.port} disconnecting...")
        with self.lock:
            if not self.is_running:
                osc_protocol_logger.info(
                    f"{self.ip_address}:{self.port} already disconnected!"
                )
                return
            self._teardown()
            if not self.osc_server._BaseServer__shutdown_request:
                self.osc_server.shutdown()
            self.osc_server = None
            self.osc_server_thread = None
        osc_protocol_logger.info(f"{self.ip_address}:{self.port} ...disconnected")

    def expect(self, message, pattern, failure_pattern=None, timeout=1.0):
        def set_response(message):
            with condition:
                result["result"] = message
                condition.notify()

        result = {"result": None}
        condition = threading.Condition()
        start_time = time.time()
        with condition:
            callback = self.register(
                pattern,
                lambda x: set_response(x),
                failure_pattern=failure_pattern,
                once=True,
            )
            self.send(message)
            while result["result"] is None:
                condition.wait(timeout)
                current_time = time.time()
                delta_time = current_time - start_time
                if timeout <= delta_time:
                    break
        self.unregister(callback)
        return result["result"]

    def register(
        self, pattern, procedure, *, failure_pattern=None, once=False
    ) -> OscCallback:
        """
        Register a callback.
        """
        callback = self._validate_callback(
            pattern, procedure, failure_pattern=failure_pattern, once=once
        )
        # Command queue prevents lock contention.
        self.command_queue.put(("add", callback))
        return callback

    def send(self, message):
        datagram = self._validate_send(message)
        try:
            self.osc_server.socket.sendto(datagram, (self.ip_address, self.port))
        except OSError:
            # print(message)
            raise

    def unregister(self, callback: OscCallback):
        """
        Unregister a callback.
        """
        # Command queue prevents lock contention.
        self.command_queue.put(("remove", callback))



"""
Enumerations.
"""

from collections.abc import Sequence

from uqbar.enums import IntEnumeration, StrictEnumeration


class AddAction(IntEnumeration):
    """
    An enumeration of scsynth node add actions.
    """

    ### CLASS VARIABLES ###

    ADD_TO_HEAD = 0
    ADD_TO_TAIL = 1
    ADD_BEFORE = 2
    ADD_AFTER = 3
    REPLACE = 4


class BinaryOperator(IntEnumeration):

    ### CLASS VARIABLES ###

    ABSOLUTE_DIFFERENCE = 38  # |a - b|
    ADDITION = 0
    AMCLIP = 40
    ATAN2 = 22
    BIT_AND = 14
    BIT_OR = 15
    BIT_XOR = 16
    CLIP2 = 42
    DIFFERENCE_OF_SQUARES = 34  # a*a - b*b
    EQUAL = 6
    EXCESS = 43
    EXPRANDRANGE = 48
    FLOAT_DIVISION = 4
    FILL = 29
    FIRST_ARG = 46
    FOLD2 = 44
    GREATEST_COMMON_DIVISOR = 18
    GREATER_THAN_OR_EQUAL = 11
    GREATER_THAN = 9
    HYPOT = 23
    HYPOTX = 24
    INTEGER_DIVISION = 3
    LEAST_COMMON_MULTIPLE = 17
    LESS_THAN_OR_EQUAL = 10
    LESS_THAN = 8
    MAXIMUM = 13
    MINIMUM = 12
    MODULO = 5
    MULTIPLICATION = 2
    NOT_EQUAL = 7
    POWER = 25
    RANDRANGE = 47
    RING1 = 30  # a * (b + 1) == a * b + a
    RING2 = 31  # a * b + a + b
    RING3 = 32  # a*a*b
    RING4 = 33  # a*a*b - a*b*b
    ROUND = 19
    ROUND_UP = 20
    SCALE_NEG = 41
    SHIFT_LEFT = 26
    SHIFT_RIGHT = 27
    SQUARE_OF_DIFFERENCE = 37  # (a - b)^2
    SQUARE_OF_SUM = 36  # (a + b)^2
    SUBTRACTION = 1
    SUM_OF_SQUARES = 35  # a*a + b*b
    THRESHOLD = 39
    TRUNCATION = 21
    UNSIGNED_SHIFT = 28
    WRAP2 = 45


class CalculationRate(IntEnumeration):
    """
    An enumeration of scsynth calculation-rates.

    ::

        >>> import supriya.synthdefs
        >>> supriya.CalculationRate.AUDIO
        CalculationRate.AUDIO

    ::

        >>> supriya.CalculationRate.from_expr("demand")
        CalculationRate.DEMAND

    """

    ### CLASS VARIABLES ###

    AUDIO = 2
    CONTROL = 1
    DEMAND = 3
    SCALAR = 0
    AR = 2
    KR = 1
    DR = 3
    IR = 0

    ### PUBLIC METHODS ###

    @classmethod
    def from_expr(cls, expr):
        """
        Gets calculation-rate.

        ::

            >>> import supriya.synthdefs
            >>> import supriya.ugens

        ::

            >>> supriya.CalculationRate.from_expr(1)
            CalculationRate.SCALAR

        ::

            >>> supriya.CalculationRate.from_expr("demand")
            CalculationRate.DEMAND

        ::

            >>> collection = []
            >>> collection.append(supriya.ugens.DC.ar(0))
            >>> collection.append(supriya.ugens.DC.kr(1))
            >>> collection.append(2.0)
            >>> supriya.CalculationRate.from_expr(collection)
            CalculationRate.AUDIO

        ::
            >>> collection = []
            >>> collection.append(supriya.ugens.DC.kr(1))
            >>> collection.append(2.0)
            >>> supriya.CalculationRate.from_expr(collection)
            CalculationRate.CONTROL

        Return calculation-rate.
        """
        import supriya.synthdefs
        import supriya.ugens

        if isinstance(expr, (int, float)) and not isinstance(expr, cls):
            return CalculationRate.SCALAR
        elif isinstance(expr, (supriya.synthdefs.OutputProxy, supriya.synthdefs.UGen)):
            return expr.calculation_rate
        elif isinstance(expr, supriya.synthdefs.Parameter):
            name = expr.parameter_rate.name
            if name == "TRIGGER":
                return CalculationRate.CONTROL
            return CalculationRate.from_expr(name)
        elif isinstance(expr, str):
            return super().from_expr(expr)
        elif isinstance(expr, Sequence):
            return max(CalculationRate.from_expr(item) for item in expr)
        elif hasattr(expr, "calculation_rate"):
            return cls.from_expr(expr.calculation_rate)
        return super().from_expr(expr)

    ### PUBLIC PROPERTIES ###

    @property
    def token(self):
        if self == CalculationRate.SCALAR:
            return "ir"
        elif self == CalculationRate.CONTROL:
            return "kr"
        elif self == CalculationRate.AUDIO:
            return "ar"
        return "new"


class DoneAction(IntEnumeration):
    """
    An enumeration of ``scsynth`` UGen "done" actions.

    ::

        >>> import supriya.synthdefs
        >>> supriya.DoneAction(2)
        DoneAction.FREE_SYNTH

    ::

        >>> supriya.DoneAction.from_expr("pause synth")
        DoneAction.PAUSE_SYNTH

    """

    ### CLASS VARIABLES ###

    NOTHING = 0
    PAUSE_SYNTH = 1
    FREE_SYNTH = 2
    FREE_SYNTH_AND_PRECEDING_NODE = 3
    FREE_SYNTH_AND_FOLLOWING_NODE = 4
    FREE_SYNTH_AND_FREEALL_PRECEDING_NODE = 5
    FREE_SYNTH_AND_FREEALL_FOLLOWING_NODE = 6
    FREE_SYNTH_AND_ALL_PRECEDING_NODES_IN_GROUP = 7
    FREE_SYNTH_AND_ALL_FOLLOWING_NODES_IN_GROUP = 8
    FREE_SYNTH_AND_PAUSE_PRECEDING_NODE = 9
    FREE_SYNTH_AND_PAUSE_FOLLOWING_NODE = 10
    FREE_SYNTH_AND_DEEPFREE_PRECEDING_NODE = 11
    FREE_SYNTH_AND_DEEPFREE_FOLLOWING_NODE = 12
    FREE_SYNTH_AND_ALL_SIBLING_NODES = 13
    FREE_SYNTH_AND_ENCLOSING_GROUP = 14


class EnvelopeShape(IntEnumeration):

    ### CLASS VARIABLES ###

    CUBED = 7
    CUSTOM = 5
    EXPONENTIAL = 2
    LINEAR = 1
    SINE = 3
    SQUARED = 6
    STEP = 0
    WELCH = 4


class HeaderFormat(IntEnumeration):
    """
    An enumeration of soundfile header formats.

    ::

        >>> supriya.HeaderFormat.AIFF
        HeaderFormat.AIFF

    ::

        >>> supriya.HeaderFormat.from_expr("wav")
        HeaderFormat.WAV

    ::

        >>> header_format = supriya.HeaderFormat.from_expr("wav")
        >>> header_format.name.lower()
        'wav'

    """

    ### CLASS VARIABLES ###

    AIFF = 0
    IRCAM = 1
    NEXT = 2
    RAW = 3
    WAV = 4


class NodeAction(IntEnumeration):

    ### CLASS VARIABLES ###

    NODE_CREATED = 0
    NODE_REMOVED = 1
    NODE_ACTIVATED = 2
    NODE_DEACTIVATED = 3
    NODE_MOVED = 4
    NODE_QUERIED = 5

    ### PUBLIC METHODS ###

    @classmethod
    def from_address(cls, address):
        addresses = {
            "/n_end": cls.NODE_REMOVED,
            "/n_go": cls.NODE_CREATED,
            "/n_info": cls.NODE_QUERIED,
            "/n_move": cls.NODE_MOVED,
            "/n_off": cls.NODE_DEACTIVATED,
            "/n_on": cls.NODE_ACTIVATED,
        }
        action = addresses[address]
        return action


class ParameterRate(IntEnumeration):
    """
    An enumeration of synthdef control rates.
    """

    ### CLASS VARIABLES ###

    AUDIO = 2
    CONTROL = 3
    SCALAR = 0
    TRIGGER = 1
    AR = 2
    KR = 3
    IR = 0
    TR = 1


class RequestId(IntEnumeration):
    """
    An enumeration of scsynth request ids.
    """

    ### CLASS VARIABLES ###

    BUFFER_ALLOCATE = 28
    BUFFER_ALLOCATE_READ = 29
    BUFFER_ALLOCATE_READ_CHANNEL = 54
    BUFFER_CLOSE = 33
    BUFFER_FILL = 37
    BUFFER_FREE = 32
    BUFFER_GENERATE = 38
    BUFFER_GET = 42
    BUFFER_GET_CONTIGUOUS = 43
    BUFFER_QUERY = 47
    BUFFER_READ = 30
    BUFFER_READ_CHANNEL = 55
    BUFFER_SET = 35
    BUFFER_SET_CONTIGUOUS = 36
    BUFFER_WRITE = 31
    BUFFER_ZERO = 34
    CLEAR_SCHEDULE = 51
    COMMAND = 4
    CONTROL_BUS_FILL = 27
    CONTROL_BUS_GET = 40
    CONTROL_BUS_GET_CONTIGUOUS = 41
    CONTROL_BUS_SET = 25
    CONTROL_BUS_SET_CONTIGUOUS = 26
    DUMP_OSC = 39
    ERROR = 58
    GROUP_DEEP_FREE = 50
    GROUP_DUMP_TREE = 56
    GROUP_FREE_ALL = 24
    GROUP_HEAD = 22
    GROUP_NEW = 21
    GROUP_QUERY_TREE = 57
    GROUP_TAIL = 23
    NODE_AFTER = 19
    NODE_BEFORE = 18
    NODE_COMMAND = 13
    NODE_FILL = 17
    NODE_FREE = 11
    NODE_MAP_TO_CONTROL_BUS = 14
    NODE_MAP_TO_AUDIO_BUS = 60
    NODE_MAP_TO_AUDIO_BUS_CONTIGUOUS = 61
    NODE_MAP_TO_CONTROL_BUS_CONTIGUOUS = 48
    NODE_ORDER = 62
    NODE_QUERY = 46
    NODE_RUN = 12
    NODE_SET = 15
    NODE_SET_CONTIGUOUS = 16
    NODE_TRACE = 10
    NOTHING = 0
    NOTIFY = 1
    PARALLEL_GROUP_NEW = 63
    QUIT = 3
    STATUS = 2
    SYNC = 52
    SYNTHDEF_FREE = 53
    SYNTHDEF_FREE_ALL = 8
    SYNTHDEF_LOAD = 6
    SYNTHDEF_LOAD_DIR = 7
    SYNTHDEF_RECEIVE = 5
    SYNTH_GET = 44
    SYNTH_GET_CONTIGUOUS = 45
    SYNTH_NEW = 9
    SYNTH_NEWARGS = 59
    SYNTH_NOID = 49
    SYNTH_QUERY = 65
    UGEN_COMMAND = 20
    VERSION = 64

    @property
    def request_name(self):
        return RequestName.from_expr(self.name)


class RequestName(StrictEnumeration):
    """
    An enumeration of scsynth request names.
    """

    ### CLASS VARIABLES ###

    BUFFER_ALLOCATE = "/b_alloc"
    BUFFER_ALLOCATE_READ = "/b_allocRead"
    BUFFER_ALLOCATE_READ_CHANNEL = "/b_allocReadChannel"
    BUFFER_CLOSE = "/b_close"
    BUFFER_FILL = "/b_fill"
    BUFFER_FREE = "/b_free"
    BUFFER_GENERATE = "/b_gen"
    BUFFER_GET = "/b_get"
    BUFFER_GET_CONTIGUOUS = "/b_getn"
    BUFFER_QUERY = "/b_query"
    BUFFER_READ = "/b_read"
    BUFFER_READ_CHANNEL = "/b_readChannel"
    BUFFER_SET = "/b_set"
    BUFFER_SET_CONTIGUOUS = "/b_setn"
    BUFFER_WRITE = "/b_write"
    BUFFER_ZERO = "/b_zero"
    CLEAR_SCHEDULE = "/clearSched"
    COMMAND = "/cmd"
    CONTROL_BUS_FILL = "/c_fill"
    CONTROL_BUS_GET = "/c_get"
    CONTROL_BUS_GET_CONTIGUOUS = "/c_getn"
    CONTROL_BUS_SET = "/c_set"
    CONTROL_BUS_SET_CONTIGUOUS = "/c_setn"
    DUMP_OSC = "/dumpOSC"
    ERROR = "/error"
    GROUP_DEEP_FREE = "/g_deepFree"
    GROUP_DUMP_TREE = "/g_dumpTree"
    GROUP_FREE_ALL = "/g_freeAll"
    GROUP_HEAD = "/g_head"
    GROUP_NEW = "/g_new"
    GROUP_QUERY_TREE = "/g_queryTree"
    GROUP_TAIL = "/g_tail"
    NODE_AFTER = "/n_after"
    NODE_BEFORE = "/n_before"
    # NODE_COMMAND = None
    NODE_FILL = "/n_fill"
    NODE_FREE = "/n_free"
    NODE_MAP_TO_AUDIO_BUS = "/n_mapa"
    NODE_MAP_TO_AUDIO_BUS_CONTIGUOUS = "/n_mapan"
    NODE_MAP_TO_CONTROL_BUS = "/n_map"
    NODE_MAP_TO_CONTROL_BUS_CONTIGUOUS = "/n_mapn"
    NODE_ORDER = "/n_order"
    NODE_QUERY = "/n_query"
    NODE_RUN = "/n_run"
    NODE_SET = "/n_set"
    NODE_SET_CONTIGUOUS = "/n_setn"
    NODE_TRACE = "/n_trace"
    # NOTHING = None
    NOTIFY = "/notify"
    PARALLEL_GROUP_NEW = "/p_new"
    QUIT = "/quit"
    STATUS = "/status"
    SYNC = "/sync"
    SYNTHDEF_FREE = "/d_free"
    # SYNTHDEF_FREE_ALL = None
    SYNTHDEF_LOAD = "/d_load"
    SYNTHDEF_LOAD_DIR = "/d_loadDir"
    SYNTHDEF_RECEIVE = "/d_recv"
    SYNTH_GET = "/s_get"
    SYNTH_GET_CONTIGUOUS = "/s_getn"
    SYNTH_NEW = "/s_new"
    SYNTH_QUERY = "/s_query"
    # SYNTH_NEWARGS = None
    SYNTH_NOID = "/s_noid"
    UGEN_COMMAND = "/u_cmd"
    VERSION = "/version"

    ### PUBLIC PROPERTIES ###

    @property
    def request_id(self):
        return RequestId.from_expr(self.name)


class SampleFormat(IntEnumeration):
    """
    An enumeration of soundfile sample formats.

    ::

        >>> supriya.SampleFormat.INT24
        SampleFormat.INT24

    ::

        >>> supriya.SampleFormat.from_expr("float")
        SampleFormat.FLOAT

    ::

        >>> sample_format = supriya.SampleFormat.INT24
        >>> sample_format.name.lower()
        'int24'

    """

    ### CLASS VARIABLES ###

    INT24 = 0
    ALAW = 1
    DOUBLE = 2
    FLOAT = 3
    INT8 = 4
    INT16 = 5
    INT32 = 6
    MULAW = 7


class SignalRange(IntEnumeration):
    """
    An enumeration of scsynth UGen signal ranges.

    ::

        >>> supriya.SignalRange.UNIPOLAR
        SignalRange.UNIPOLAR

    ::

        >>> supriya.SignalRange.from_expr("bipolar")
        SignalRange.BIPOLAR

    """

    ### CLASS VARIABLES ###

    UNIPOLAR = 0
    BIPOLAR = 1


class UnaryOperator(IntEnumeration):

    ### CLASS VARIABLES ###

    ABSOLUTE_VALUE = 5
    AMPLITUDE_TO_DB = 22
    ARCCOS = 32
    ARCSIN = 31
    ARCTAN = 33
    AS_FLOAT = 6
    AS_INT = 7
    BILINRAND = 40
    BIT_NOT = 4
    CEILING = 8
    COIN = 44
    COS = 29
    COSH = 35
    CUBED = 13
    DB_TO_AMPLITUDE = 21
    DIGIT_VALUE = 45
    DISTORT = 42
    EXPONENTIAL = 15
    FLOOR = 9
    FRACTIONAL_PART = 10
    HZ_TO_MIDI = 18
    HZ_TO_OCTAVE = 24
    HANNING_WINDOW = 49
    IS_NIL = 2
    LINRAND = 39
    LOG = 25
    LOG10 = 27
    LOG2 = 26
    MIDI_TO_HZ = 17
    SEMITONES_TO_RATIO = 19
    NEGATIVE = 0
    NOT = 1
    NOT_NIL = 3
    OCTAVE_TO_HZ = 23
    RAMP = 52
    RAND = 37
    RAND2 = 38
    RATIO_TO_SEMITONES = 20
    RECIPROCAL = 16
    RECTANGLE_WINDOW = 48
    S_CURVE = 53
    SIGN = 11
    SILENCE = 46
    SIN = 28
    SINH = 34
    SOFTCLIP = 43
    SQUARE_ROOT = 14
    SQUARED = 12
    SUM3RAND = 41
    TAN = 30
    TANH = 36
    THRU = 47
    TRIANGLE_WINDOW = 51
    WELCH_WINDOW = 50


class Unit(IntEnumeration):

    ### CLASS VARIABLES ###

    UNDEFINED = 0
    DECIBELS = 1
    AMPLITUDE = 2
    SECONDS = 3
    MILLISECONDS = 4
    HERTZ = 5
    SEMITONES = 6
